/*
	File:		FileName.h

	Function:	Provides a file spec + related ops.
				Assumed representation is path:file:extension.

	Author:		Andrew Willmott, (c) 1996-1999
*/

#ifndef __FileName__
#define __FileName__

#include "cl/String.h"
#include "cl/Flags.h"
#include <stdio.h>

class FileName
/// Class for manipulating filenames and paths
{
public:
	FileName() : path(), name(), extension(), flags() {};

	FileName	&SetPath(StrConst path);
				///< Set by the full pathname.
	FileName	&SetRelPath(StrConst path);
				/**< Set by relative path. 

					If the path starts with '/', the effect is the same
					as with SetPath, otherwise the path is appended to
					the current directory path.
				*/
	String		GetPath() const;
				///< Get the full pathname

	StrConst	GetFile() const
				///  Get the name of the file.
				{ return(name); };
	StrConst	GetExtension() const
				///  Get the file's current extension
				{ return(extension); };
	StrConst	GetDir() const
				///  Get the pathname of the parent directory of the file
		        { return(path); };

	FileName	&SetFile(StrConst file)
				///  Set the file's name
				{ name = file; return(SELF); };
	FileName	&SetExtension(StrConst ext)
				///  Set the file's extension
				{ extension = ext; return(SELF); };
	FileName	&SetDir(StrConst dir)
				///  Set the file's parent directory
				{ path = dir; return(SELF); };

	Void		AddExtension(StrConst ext);
				/**< The current extension is added to the name, and
					'ext' becomes the new file extension. */
	Void		RemoveExtension();
				/**< Strips the current extension, and replaces it with the previous 
					 extension, if one exists. */

	Bool		Exists() const;			///< Does the file exist?
	Bool		IsReadable() const;		///< Is it readable?
	Bool		IsWritable() const;		///< Is it writable?
	Long		GetTimeStamp() const;	///< Return file's timestamp 

	FILE		*FOpen(StrConst permissions) const
				///  Return filehandle to the file
				{ return(fopen(GetPath(), permissions)); };

	Int			FindFileExtension(StrConst extensions[],
					Bool allowCompressed = true);
				/**< Search for file with the given list of acceptable
					 extensions. */
	Void		MakeUnique();
				/**< Make the filename unique.
					If a file of the current name already exists, it is
					renamed 'myname-1.ext'. If that exists, it is
					renamed 'myname-2.ext', and so on until a file with
					that name doesn't already exist
				*/
	Int			DecompressSetup();
				///< Set up for decompression.
	Void		DecompressCleanup();
				///< Clean up after decompression.
	Void		AddCompExt();
				///< Add compression extension to filename
				
	Flags32		flags;		///< file info

protected:
	String		path;		///< full path to the parent directory
	String		name;		///< current file name.
	String		extension;	///< file's extension.
};

String TempPath();

enum FN_Flags
{
	FN_Gzipped = 0x0001,
	FN_Bz2ed = 0x0002,
	FN_Compress = 0x0004,

	FN_IsCompressed = 0x000F
};

enum FN_Err
{
	kFileNotFound = -1,
	kBadExtension = -2
};

#endif
