/*
	File:		FCInfo.cc

	Function:	Test program for face cluster bounds

	Author:		Andrew Willmott

	Notes:		
*/

#include <stdlib.h>
#include "arg.h"
#include "Scene.h"
#include "MRModel.h"
#include "Samples.h"
#include "VecUtil.h"
#include "MatUtil.h"
#include "Quaternion.h"
#include "Timer.h"
#include "Geometry.h"

#include "VolBounds.h"
#include "VolClus.h"
#include "MER.h"

typedef Array<Byte> ByteArray;

static VectorList		tDirSamples;
const Int				kNumTesselations = 8;

#define OBB_OUT if (0) cerr
#define MER_OUT if (0) cerr


// --- bounding box calculation stuff ---------------------------


#ifdef FCI_QHULL
extern "C"
{
#include "qhull.h"
char qh_version[] = "GCL";
}

Int FindConvexHull(PointList &points)
// use qhull lib to find convex hull of points
{
	Int		error;

	error = qh_new_qhull(3, points.NumItems(), (coordT*) points.Ref(), 
		false, "qhull Pp", 0, stderr);

	if (!error)
	{
	    vertexT *vertex;
	    Int num = 0;
	    
		FORALLvertices
		{
			points[num++] = *((Vec3d*) vertex->point);
		}
		points.SetSize(num);
	}
//	else
//		_Error("qhull failed");

	Int stupid1, stupid2;
	qh_freeqhull(!qh_ALL);
	qh_memfreeshort(&stupid1, &stupid2);

	return(error);
}
#endif

Void CalcBoundingBoxes(MRModel *model, Int ic, VecTrans axis, 
	Point &min, Point &max)
{
	Int				i;

	if (model->IsLeafCluster(ic))
	{
		Point	v;
		
		for (i = 0; i < 3; i++)
		{
			// rotate v to be in bbox coords
			v = axis * (*model->points)[model->faces[ic].v[i]];
			FindMaxElts(max, v, max);
			FindMinElts(min, v, min);
		}
	}
	else
	{
		FaceCluster 	&clus = model->Cluster(ic);

		CalcBoundingBoxes(model, clus.child[0], axis, min, max);
		CalcBoundingBoxes(model, clus.child[1], axis, min, max);
	}
}

Void CalcBoundingBoxes(MRModel *model)
{
	Int			i;
	VecTrans	axes;
	
	for (i = model->faces.NumItems(); i < model->faces.NumItems() + model->clusters.NumItems(); i++)
	{
		FaceCluster &clus = model->Cluster(i);

		clus.min.MakeBlock(HUGE_VAL);
		clus.max.MakeBlock(-HUGE_VAL);

		clus.GetAxes(axes);
		CalcBoundingBoxes(model, i, axes, clus.min, clus.max);
	}
}

Void CH_CalcBoundingBoxes(MRModel *model, Int ic, PointList &points, Int depth = 0)
{
	Int				i;
	Vector			v;
	VecTrans		axis;
	
	if (model->IsLeafCluster(ic))
	{
		points.Clear();
		points.Append((*model->points)[model->faces[ic].v[0]]);
		points.Append((*model->points)[model->faces[ic].v[1]]);
		points.Append((*model->points)[model->faces[ic].v[2]]);
	}
	else
	{
		FaceCluster 	&clus = model->Cluster(ic);
		PointList 		newPoints;

		CH_CalcBoundingBoxes(model, clus.child[0], points, depth + 1);
		CH_CalcBoundingBoxes(model, clus.child[1], newPoints, depth + 1);
		points.Append(newPoints);
#ifdef FCI_QHULL
		if (points.NumItems() > 64)
			FindConvexHull(points);
#endif

		clus.min.MakeBlock(HUGE_VAL);
		clus.max.MakeBlock(-HUGE_VAL);
		clus.GetAxes(axis);
		
		for (i = 0; i < points.NumItems(); i++)
		{
			// rotate v to be in bbox coords
			v = axis * points[i];
			FindMaxElts(clus.max, v, clus.max);
			FindMinElts(clus.min, v, clus.min);
		}
	}
}

Void CalcBoundingBoxesVoB(MRModel *model, Int ic, VolBounds &vb)
// calculate bounding boxes using the VolBounds class
{
	Int				i;
	VecTrans		axes;
	
	OBB_OUT << "process cluster " << ic << endl;
	if (model->IsLeafCluster(ic))
	{
		Int			*triIdx = model->faces[ic].v;
		PointList	&points = *model->points;

		OBB_OUT << "LEAF" << endl;
		Assert(!vl_is_finite(vb.max[0]), "Bad initial bounds at leaf");
		OBB_OUT << "tri " << points[triIdx[0]] << points[triIdx[1]] << points[triIdx[2]] << endl;
		vb.AddPoint(points[triIdx[0]]);
		vb.AddPoint(points[triIdx[1]]);
		vb.AddPoint(points[triIdx[2]]);
	}
	else
	{
		FaceCluster &clus = model->Cluster(ic);
		VolBounds	vb2;

		CalcBoundingBoxesVoB(model, clus.child[0], vb);
		CalcBoundingBoxesVoB(model, clus.child[1], vb2);
		vb.MergeBounds(vb2);

		clus.GetAxes(axes);
		vb.FindBoundingBox(axes, clus.min, clus.max);
		clus.SetAxes(axes);
	}
}

/*

	ooobbbbboooo: inverted

	bbbooooobbbb: plain

	problem: both inverted.

 */

Void MergeContours(IndexList &con1, IndexList &con2, ByteArray &marks)
/// Merge two model contours known to share a boundary.
{
	Int					i, lastIdx, start1, start2;
	Bool				marked, lastMarked, active1, active2;
	static	IndexList	newContour;

	enum { fMark1 = 1, fMark2 = 2 };

	MER_OUT << "merging " << con1 << con2 << endl;	
	// tag contours so we can locate boundary
	for (i = 0; i < con1.NumItems(); i++)
	{
		Assert(marks[con1[i]] == 0, "array still marked");
		marks[con1[i]] |= fMark1;
	}
	for (i = 0; i < con2.NumItems(); i++)
		marks[con2[i]] |= fMark2;

	newContour.SetSize(0);
	start1 = start2 = -1;
	
	// add contour1 vertices not on the boundary, and the
	// first boundary vertex.

	// we start outputting vertices when we hit a marked->unmarked transition
	// we stop when we've output the first marked vertex.
	// so, repeat while we're either not active, or the last vertex wasn't
	// marked.
	active1 = false;
	lastIdx = con1.Last();
	lastMarked = marks[lastIdx] & fMark2;
	for (i = 0; (!active1 || !lastMarked); i++)
	{
		if (i == con1.NumItems())
		{
			if (active1)
				i = 0;
			else
				break;
		}
		
		marked = marks[con1[i]] & fMark2;

		if (lastMarked && !marked)
		{
			active1 = true;
			start1 = lastIdx;
		}
		if (active1)
			newContour.Append(con1[i]);
		
		lastMarked = marked;
		lastIdx = con1[i];
	}

	// Do the same for second contour
	active2 = false;
	lastIdx = con2.Last();
	lastMarked = marks[lastIdx] & fMark1;
	for (i = 0; (!active2 || !lastMarked); i++)
	{
		if (i == con2.NumItems())
		{
			if (active2)
				i = 0;
			else
				break;
		}

		marked = marks[con2[i]] & fMark1;

		if (lastMarked && !marked)
		{
			active2 = true;
			start2 = lastIdx;
		}
		if (active2)
			newContour.Append(con2[i]);
		
		lastMarked = marked;
		lastIdx = con2[i];
	}

	// fix up case where one contour is all boundary.

	if (!active1)
		newContour.Append(start2);
	else if (!active2)
		newContour.Append(start1);
	if (!active1 || !active2)
	{
		cout << "warning: surrounded contour" << endl;
		cout << con1 << con2 << endl;
		cout << newContour << endl;
		cout << active1 << ' ' << active2 << endl;
		cout << start1 << ' ' << start2 << endl;
	}
	if (!active1 && !active2)
		_Error("alg fucked");

	// clean up
	for (i = 0; i < con1.NumItems(); i++)
		marks[con1[i]] = 0;
	for (i = 0; i < con2.NumItems(); i++)
		marks[con2[i]] = 0;

	// XXX many passes through the contours! how can we do better?

	con1 = newContour;
	MER_OUT << "result = " << con1 << endl;
}

Void FindBestFitAxes(MRModel *model, Int ic, IndexList &contour, ByteArray &marks)
/** Find planar axes via MER of the tree rooted at 'ic'.

	Assumes 'marks' has been set to the size of the model's points array,
	and cleared to zero.
*/
{
	Int				i;
	VecTrans		axes;
	PointList		&points = *model->points;
	
	if (model->IsLeafCluster(ic))
	{
		Int			*triIdx = model->faces[ic].v;

		contour.SetSize(3);
		contour[0] = triIdx[0];
		contour[1] = triIdx[1];
		contour[2] = triIdx[2];
	}
	else
	{
		FaceCluster         &clus = model->Cluster(ic);
		IndexList           contour2;
		Vector				v0, v1;
		Coord               merAxis;
		Int					minIdx[2], maxIdx[2];
		static CoordList    poly;
		
		FindBestFitAxes(model, clus.child[0], contour, marks);
		FindBestFitAxes(model, clus.child[1], contour2, marks);

		MER_OUT << "--- ic " << ic << endl;
		MER_OUT << "merging " << contour.NumItems() << " and "  
			<<contour2.NumItems() << " point contours" << endl;

		MergeContours(contour, contour2, marks);
		MER_OUT << "new contour = " << contour.NumItems() << endl;

		if (len(clus.areaNormal) / clus.totArea < 0.5)
			return;
			
		// find MER
		clus.GetAxes(axes);
		poly.SetSize(contour.NumItems());
		MER_OUT << "normal = " << axes[2] << endl;
		v0 = norm(FindOrthoVector(axes[2]));
		v1 = cross(axes[2], v0);
		MER_OUT << "in axes: " << v0 << v1 << endl;
		MER_OUT << len(v0) << ' ' << len(v1) << ' ' << len(axes[2]) << endl;

		// project contour to 2D
		for (i = 0; i < contour.NumItems(); i++)
			poly[i] = Coord(dot(points[contour[i]], v0),
						dot(points[contour[i]], v1));
	
		MER_OUT << "poly = " << poly << endl;
		if (0)
			poly.WriteFile(String().Printf("fc-%d.poly", ic));
		MER_OUT << "wrote poly" << endl;

		// find the minimal enclosing rectangle of the contour
		FindConvexHull(poly, minIdx, maxIdx);
		MER_OUT << "hull = " << poly << endl;
		MinEnclRect(poly, minIdx, maxIdx, merAxis);

		MER_OUT << "axis = " << merAxis << endl;
			
		// reconstruct 3d axes
		axes[0] = v0 * merAxis[0] + v1 * merAxis[1];
		merAxis = -cross(merAxis);
		axes[1] = v0 * merAxis[0] + v1 * merAxis[1]; 
		
		clus.SetAxes(axes);
	}
}




// --------------------------------------------------------------


Int FindClusterFaces(MRModel *model, Int ic, FaceIdxArray &triList)
{
	Int				i;

	if (model->IsLeafCluster(ic))
	{
		triList.Append(model->faces[ic]);

		return(1);	
	}
	else
	{
		FaceCluster 	&clus = model->Cluster(ic);

		return(
			FindClusterFaces(model, clus.child[0], triList) +
			FindClusterFaces(model, clus.child[1], triList)
		);
	}
}

Int PrintFacesExcludingCluster(MRModel *model, Int ic, Int extIC)
{
	Int				i;

	if (model->IsLeafCluster(ic))
	{
		cout << "f " 
			<< model->faces[ic].v[0] + 1 << ' '
			<< model->faces[ic].v[1] + 1 << ' '
			<< model->faces[ic].v[2] + 1 << ' '
			<< endl;
		return(1);	
	}
	else if (ic != extIC)
	{
		FaceCluster 	&clus = model->Cluster(ic);

		return(
			PrintFacesExcludingCluster(model, clus.child[0], extIC) +
			PrintFacesExcludingCluster(model, clus.child[1], extIC)
		);
	}
}

Void PrintClusterAsObj(MRModel *model, Int ic, Bool showRest)
{
	Int				i, j, facesExtracted;
	FaceIdxArray	triList;
	
	facesExtracted = FindClusterFaces(model, ic, triList);
	cout << "mtllib clus.mtl" << endl;

	if (showRest)
	{
		for (i = 0; i < model->points->NumItems(); i++)
			cout << "v " 
				<< (*model->points)[i][0] << ' '
				<< (*model->points)[i][1] << ' '
				<< (*model->points)[i][2] << ' '
				<< endl;

		cout << "usemtl object" << endl;
		for (i = 0; i < model->rootClusters.NumItems(); i++)
			PrintFacesExcludingCluster(model, model->rootClusters[i], ic);

		cout << "usemtl cluster" << endl;
		for (i = 0; i < triList.NumItems(); i++)
			cout << "f " 
				<< triList[i].v[0] + 1 << ' '
				<< triList[i].v[1] + 1 << ' '
				<< triList[i].v[2] + 1 << ' '
				<< endl;
		
		return;
	}

	IndexList	vertexMap;
	Int			newVtx = 0;
	
	cout << "usemtl cluster" << endl;
	vertexMap.SetSize(model->points->NumItems());

	for (i = 0; i < vertexMap.NumItems(); i++)
		vertexMap[i] = -1;

	for (i = 0; i < triList.NumItems(); i++)
		for (j = 0; j < 3; j++)
		{
			Int		idx = triList[i].v[j];
			
			if (vertexMap[idx] < 0)
			{
				vertexMap[idx] = newVtx++;

				cout << "v "
					<< (*model->points)[idx][0] << ' '
					<< (*model->points)[idx][1] << ' '
					<< (*model->points)[idx][2] << ' '
					<< endl;
			}
			triList[i].v[j] = vertexMap[idx];
		}

	for (i = 0; i < triList.NumItems(); i++)
		cout << "f "
			<< triList[i].v[0] + 1 << ' '
			<< triList[i].v[1] + 1 << ' '
			<< triList[i].v[2] + 1 << ' '
			<< endl;

	cerr << "cluster has " << facesExtracted << " faces" << endl;
}

// --- Sphere stuff ------------------------------------------

PointList MySpherePoints()
{
	Int		i, j;
	GCLReal	theta, dtheta, phi, dphi, z;
	PointList	list;
	
	list.Append(Point(0.0, 0.0, 1.0));

	phi = 2.0 * vl_pi / kNumTesselations;
	dphi = phi;
	dtheta = dphi;

	for (j = 0; j < (kNumTesselations / 2) - 1; j++)
	{
		z = cos(phi);
		if (j & 1)
			theta = dtheta / 2.0;
		else
			theta = 0.0;
		
		for (i = 0; i < kNumTesselations; i++)
		{
			list.Append(Point(sin(phi) * cos(theta),
						  sin(phi) * sin(theta),
						  z));
			theta += dtheta;
		}
		phi += dphi;
	}	
	
	list.Append(Point(0.0, 0.0, -1.0));

	return(list);
}

Int ReverseSphereDir(Int i)
{
	const Int levels = (kNumTesselations / 2 - 1);
	const Int strips = kNumTesselations;
	const Int last = levels * strips + 1;
	Int		j, k;
	
	if (i == 0)
		return(last);
	else if (i == last)
		return(0);

	i--;
	j = i / strips;
	k = i - j * strips;

	j = levels - j - 1;
	k = (k + strips/2) % strips;
		
	return(strips * j + k + 1);
}

Void GetSphereFaces(IndexList &triIndices)
{
	Int			i, j, p, offset;
	Int			triangle[3], square[4];

	p = kNumTesselations;

	for (i = 1; i <= kNumTesselations; i++)
		// Step around the sphere doing vertical strips from pole to pole... 
	{
		// Make a face on the top of the sphere 
		triIndices.Append(0);
		triIndices.Append(p);
		triIndices.Append(i);

		// Make the faces between rings of vertices... 
		offset = 0;
		for (j = 0; j < (kNumTesselations / 2) - 2; j++)
		{
			// Make one such face 
			square[0] = i + offset;
			square[1] = p + offset;
			offset += kNumTesselations;
			square[2] = p + offset;
			square[3] = i + offset;

			if (j & 1)
			{
				triIndices.Append(square[0]);
				triIndices.Append(square[1]);
				triIndices.Append(square[3]);
				triIndices.Append(square[1]);
				triIndices.Append(square[2]);
				triIndices.Append(square[3]);
			}
			else
			{
				triIndices.Append(square[0]);
				triIndices.Append(square[1]);
				triIndices.Append(square[2]);
				triIndices.Append(square[2]);
				triIndices.Append(square[3]);
				triIndices.Append(square[0]);
			}
		}

		// Make a face on the bottom of the sphere 
		triIndices.Append(offset + kNumTesselations + 1); // last vertex
		triIndices.Append(offset + i);
		triIndices.Append(offset + p);

		p = i;
	}
}

static Void PrintSphereFaces()
{
	IndexList 	triIndices;
	Int			i;
	
	GetSphereFaces(triIndices);

	for (i = 0; i < triIndices.NumItems(); i += 3)
		cout << "f " 
			 << triIndices[i] + 1 << ' '
			 << triIndices[i + 1] + 1 << ' '
			 << triIndices[i + 2] + 1 << endl;
}

Void WriteObjTable(Vecd &table)
{
	Int				i;
	
#ifdef OLD
	cout << "v 0 0 0" << endl;
#endif
	for (i = 0; i < table.Elts(); i++)
	{
		Point p = tDirSamples[i] * table[i];

		cout << "v " << p[0] << ' ' << p[1] << ' ' << p[2] << endl;
	}

#ifdef OLD
	for (i = 0; i < table.Elts(); i++)
	{
		cout << "f 1 1 " << i + 2 << endl;
	}
#else
	PrintSphereFaces();
#endif
}

Void FindPAsFromSumAreaNormal(MRModel *model, Int ic, Vecd &table)
{
	FaceCluster 	&clus = model->Cluster(ic);
	Int				i;
	
	for (i = 0; i < table.Elts(); i++)
	{
		Point p = tDirSamples[i];
		Double dp = dot(p, clus.areaNormal);

		if (dp > 0.0)
			table[i] = dp;
		else
			table[i] = 0.0;
	}
}

Void FindDPs(MRModel *model, Int ic, Vecd &table)
// find cosine between cluster areaNormal direction and table direction
{
	FaceCluster 	&clus = model->Cluster(ic);
	Int				i;
	
	for (i = 0; i < table.Elts(); i++)
	{
		Point p = tDirSamples[i];
		Double dp = dot(p, clus.areaNormal);

		dp /= len(clus.areaNormal);
		table[i] = dp;
	}
}

Void FindActualPAs(MRModel *model, Int ic, Vecd &table)
{
	Int				i;
	
	if (model->IsLeafCluster(ic))
		FindPAsFromSumAreaNormal(model, ic, table);
	else
	{
		FaceCluster 	&clus = model->Cluster(ic);
		Vecd			tableL(table.Elts()), tableR(table.Elts());

		FindActualPAs(model, clus.child[0], tableL);
		FindActualPAs(model, clus.child[1], tableR);
		table = tableL + tableR;
	}
}

Void CalcAxisAlignedBoxPAs(MRModel *model, Int ic, Double box[6])
// calculate projected area of cluster in 6 unit directions
{
	Int				i;
	
	if (model->IsLeafCluster(ic))
	{
		Vector areaNormal = model->FaceAreaNormal(ic);
		
		if (areaNormal[0] > 0.0)
			box[0] += areaNormal[0];
		else
			box[3] -= areaNormal[0];
		if (areaNormal[1] > 0.0)
			box[1] += areaNormal[1];
		else
			box[4] -= areaNormal[1];
		if (areaNormal[2] > 0.0)
			box[2] += areaNormal[2];
		else
			box[5] -= areaNormal[2];
	}
	else
	{
		FaceCluster 	&clus = model->Cluster(ic);

		CalcAxisAlignedBoxPAs(model, clus.child[0], box);
		CalcAxisAlignedBoxPAs(model, clus.child[1], box);
	}
}

Void FindPAsFromAxisAlignedBox(Double box[6], Vecd &table)
// find projected areas from axis-aligned box.
{	
	Int		i;
	
	for (i = 0; i < table.Elts(); i++)
	{
		Point p = tDirSamples[i];

		table[i] = 0.0;
		if (p[0] > 0.0)
			table[i] += p[0] * box[0];
		else
			table[i] -= p[0] * box[3];
		if (p[1] > 0.0)
			table[i] += p[1] * box[1];
		else
			table[i] -= p[1] * box[4];
		if (p[2] > 0.0)
			table[i] += p[2] * box[2];
		else
			table[i] -= p[2] * box[5];
	}
}


Int FindDepth(MRModel *model, Int ic)
// how many levels down is cluster ic?
{
	Int		i;
	Int		levels = 0;
	
	for (i = ic + 1; i < model->faces.NumItems() + model->clusters.NumItems(); i++)
		if (model->Cluster(i).child[0] == ic || 
			model->Cluster(i).child[1] == ic)
		{
			levels++;
			ic = i;
		}

	return(levels);
}

// --- projected area bounds using cluster bounding boxes -----------------

Void InitBox(Double box[6])
{
	Int		i;

	for (i = 0; i < 6; i++)
		box[i] = 0.0;
}

Void CalcBoundingBoxPAs(MRModel *model, Int ic, Double box[6], const VecTrans &axis)
{
	Int				i;
	
	if (model->IsLeafCluster(ic))
	{
		Vector pa = axis * model->FaceAreaNormal(ic);

		if (pa[0] > 0.0)
			box[0] += pa[0];
		else
			box[3] -= pa[0];
		if (pa[1] > 0.0)
			box[1] += pa[1];
		else
			box[4] -= pa[1];
		if (pa[2] > 0.0)
			box[2] += pa[2];
		else
			box[5] -= pa[2];
	}
	else
	{
		FaceCluster 	&clus = model->Cluster(ic);

		CalcBoundingBoxPAs(model, clus.child[0], box, axis);
		CalcBoundingBoxPAs(model, clus.child[1], box, axis);
	}
}

Void CalcBoundingBoxPAs(MRModel *model, Int ic, Double box[6])
{
	FaceCluster 	&clus = model->Cluster(ic);
	Int				i;
	VecTrans		axes;
	
	clus.GetAxes(axes);
	InitBox(box);

	CalcBoundingBoxPAs(model, ic, box, axes);
}

Void CalcBoundingBoxPAs(MRModel *model)
{
	Int			i, j;
	VecTrans	axes;
	Double		box[6];
	Vector		diff;

	for (i = model->faces.NumItems(); i < model->faces.NumItems() + model->clusters.NumItems(); i++)
	{
		FaceCluster &clus = model->Cluster(i);

		diff = clus.max - clus.min;
		
		InitBox(box);
		clus.GetAxes(axes);

		CalcBoundingBoxPAs(model, i, box, axes);

		// find as fraction of upper bound: box's side area.
		box[0] = Min(box[0] / (diff[1] * diff[2]), 1.0);
		box[3] = Min(box[3] / (diff[1] * diff[2]), 1.0);
		box[1] = Min(box[1] / (diff[0] * diff[2]), 1.0);
		box[4] = Min(box[4] / (diff[0] * diff[2]), 1.0);
		box[2] = Min(box[2] / (diff[0] * diff[1]), 1.0);
		box[5] = Min(box[5] / (diff[0] * diff[1]), 1.0);

		for (j = 0; j < 6; j++)
			clus.sideArea[j] = Byte(255.0 * box[j]);
	}
}


Void FindPAsFromBoundingBox(MRModel *model, Int ic, Vecd &table, Double box[6])
// if box == 0 (no area samples), side areas of bounding box are used.
{
	FaceCluster 	&clus = model->Cluster(ic);
	Int				i;
	Vector			diff, lens, area, res, dps;
	VecTrans		axis;
	
	clus.GetAxes(axis);
	
	lens[0] = len(axis[0]);
	lens[1] = len(axis[1]);
	lens[2] = len(axis[2]);
	diff = clus.max - clus.min;
	area[0] = (diff[1] * lens[1] * diff[2] * lens[2]) / lens[0];
	area[1] = (diff[2] * lens[2] * diff[0] * lens[0]) / lens[1];
	area[2] = (diff[0] * lens[0] * diff[1] * lens[1]) / lens[2];

	// the maximum side area for an optimal bounding box
	// should be 2 * totalArea. PCA boxes aren't optimal,
	// however.
	Expect(area[0] <= 3.0 * clus.totArea, "bad axis area");
	Expect(area[1] <= 3.0 * clus.totArea, "bad axis area");
	Expect(area[2] <= 3.0 * clus.totArea, "bad axis area");

	for (i = 0; i < table.Elts(); i++)
	{
		Point p = tDirSamples[i];

		if (!box)
		{
			// use side areas
			res = (axis * p) * area;
			table[i] = abs(res[0]) + abs(res[1]) + abs(res[2]);
		}
		else
		{
			dps = (axis * p) / lens;
			table[i] = 0.0;
			if (dps[0] > 0.0)
				table[i] += dps[0] * box[0];
			else
				table[i] -= dps[0] * box[3];
			if (dps[1] > 0.0)
				table[i] += dps[1] * box[1];
			else
				table[i] -= dps[1] * box[4];
			if (dps[2] > 0.0)
				table[i] += dps[2] * box[2];
			else
				table[i] -= dps[2] * box[5];
		}
	}
}

Double MaxError(const Vecd &diff)
{
	Int		i;
	Double	maxErr = 0.0;
	
	for (i = 0; i < diff.Elts(); i++)
	{
		Double x = abs(diff[i]);

		if (x > maxErr) 
			maxErr = x;
	}

	return(maxErr);
}

Double RMSError(const Vecd &diff)
{
	return(sqrt(dot(diff, diff) / diff.Elts()));
}

// --- Projected Area Cube routines -------------------------------------------

// for tracking projected area over a range of (fixed) directions

const Int kPACubeSide = 8;
typedef Float PACube[6][kPACubeSide][kPACubeSide];

Void InitPACube(MRModel *model, Int fc, PACube paCube)
{
	Int		i, j, k;

	for (i = 0; i < 6; i++)
		for (j = 0; j < kPACubeSide; j++)
			for (k = 0; k < kPACubeSide; k++)
				paCube[i][j][k] = 0.0;
}

inline Double ToCubeIdx(Int x)
{
	return(2.0 * (x + 0.5) / kPACubeSide - 1.0);
}

inline Int FromCubeIdx(Double x)
{
	Int r = Int(0.5 * (x + 1.0) * kPACubeSide);

	if (r == kPACubeSide)
		r--;

	return(r);
}

Void FindPACube(MRModel *model, Int ic, PACube paCube)
// add S into paCube.
{
	FaceCluster 	&clus = model->Cluster(ic);
	Int				i, j, k, d0, d1, d2;
	Point			p;
	Double			dp;
		
	for (i = 0; i < 6; i++)
	{
		d0 = i % 3;
		d1 = (d0 + 1) % 3;
		d2 = (d1 + 1) % 3;

		// pick major axis
		if (i == d0)
			p[d0] = 1.0;
		else
			p[d0] = -1.0;

		for (j = 0; j < kPACubeSide; j++)
			for (k = 0; k < kPACubeSide; k++)
			{
				p[d1] = ToCubeIdx(j);
				p[d2] = ToCubeIdx(k);

				dp = dot(norm(p), clus.areaNormal);
				if (dp > 0.0)
					paCube[i][j][k] += dp;
			}
	}
}

// AAA
inline Vec3d abs(const Vec3d &v)
{
	Vec3d result;

	result[0] = abs(v[0]);
	result[1] = abs(v[1]);
	result[2] = abs(v[2]);
	
	return(result);
}

Float FindPAFromCube(PACube paCube, Point &p)
{
	Int				i, j, k, d0, d1, d2;
	Point 			pa = abs(p);
	
	d0 = MaxEltIndex(pa);
	d1 = (d0 + 1) % 3;
	d2 = (d1 + 1) % 3;
	if (d0 >= 0)
		i = d0;
	else
		i = d0 + 3;

	j = FromCubeIdx(p[d1] / pa[d0]);
	k = FromCubeIdx(p[d2] / pa[d0]);

	return(paCube[i][j][k]);
}

// --- stats --------------------------------------------------

Void FindVolStats(MRModel *model)
{
	Int		i, n;
	Double	sumVol, sumModVol, modVol, vol;
	Double	sumArea, area;

	sumVol = sumModVol = modVol = 0.0;
	sumArea = 0.0;
	
	n = 0;
	
	// XXX
	for (i = model->faces.NumItems(); i < model->faces.NumItems() + model->clusters.NumItems(); i++)
	{
		FaceCluster &fc = model->Cluster(i);

		vol = BoxVol(fc.min, fc.max);
		area = BoxArea(fc.min, fc.max);

		// update model volume by adding in the vol of the new cluster
		modVol += vol;
		// and subtract off its children's volumes
		if (!model->IsLeafCluster(fc.child[0]))
		{
			FaceCluster &fcc = model->Cluster(fc.child[0]);
			modVol -= BoxVol(fcc.min, fcc.max);
		}
		if (!model->IsLeafCluster(fc.child[1]))
		{
			FaceCluster &fcc = model->Cluster(fc.child[1]);
			modVol -= BoxVol(fcc.min, fcc.max);
		}
		
		sumVol += vol;
		sumArea += area;
		sumModVol += modVol;
		
		n++;
	}

	cout << "average cluster volume: " << sumVol / n << endl;
	cout << "average cluster surface area: " << sumArea / n << endl;
	cout << "average model volume: " << sumModVol / n << endl;
}

// --- main stuff ---------------------------------------------

Int ToClusterRef(MRModel *model, StrConst s)
// translate cluster reference into cluster ID
{
	Int a = s.FindChar(':');
	if (a >= 0)
	{
		Int		root = atoi(s.Prefix(a)), i, id, lastID = -1;
		String	code = s.Suffix(-a - 1);

		if (root >= 0 && root < model->rootClusters.NumItems())
			id = model->rootClusters[root];
		else
		{
			cerr << "Bad root cluster: " << root << endl;
			exit(-1);
		}

		for (i = 0; i < code.Length(); i++)
		{
			if (model->IsLeafCluster(id))
			{	
				cerr << "Warning: hit bottom of tree" << endl;
				id = lastID;
				break;
			}
			lastID = id;
			if (code[i] == '0')
				id = model->Cluster(id).child[0];
			else
				id = model->Cluster(id).child[1];
		}
		
		return(id);
	}
	else
		return(atoi(s));
}

static MRModel *gCSCModel = 0;

Int ClusSizeCompare(const Void *c1, const Void *c2)
{
	Int		i1 = *((Int*) c1);
	Int		i2 = *((Int*) c2);
	Double	size1, size2;

	if (gCSCModel->IsLeafCluster(i1))
		size1 = len(gCSCModel->FaceAreaNormal(i1));
	else
		size1 = gCSCModel->Cluster(i1).totArea;

	if (gCSCModel->IsLeafCluster(i2))
		size2 = len(gCSCModel->FaceAreaNormal(i2));
	else
		size2 = gCSCModel->Cluster(i2).totArea;
	
	if (size1 < size2)
		return(1);
	else if (size1 == size2)
		return(0);
	else
		return(-1);
}

Void SortRootClusters(MRModel *model)
{
	gCSCModel = model;
	qsort((Void *) model->rootClusters.Ref(), model->rootClusters.NumItems(),
		sizeof(Int), ClusSizeCompare);
}

#ifdef FCI_SAMPLES
Void InitSamples(Int numSamples)
{
	Int		i;
	
	tDirSamples.SetSize(numSamples);
	for (i = 0; i < tDirSamples.NumItems(); i++)
		tDirSamples[i] = *((Point*) tSphereSamples[i]);
}
#endif

Int main(Int argc, Char *argv[])
{
	MRModel		*model = new MRModel;
	Double		box1[6], box2[6];
	Int			i, levels, allStats;
	StrConst	clusStr = 0, clusExtStr = 0;
	Int			clus, clusExt, clusTable = -1, showRest, volClus, metricStats;
	Int			calcOBB = -1, setAA;
	StrConst	filename;
	FileName	modelFile;
	Arg_form	*arg_format;

//	InitSamples(256);

	tDirSamples = MySpherePoints();
	
	const Int	samples = tDirSamples.NumItems();
	Vecd		table1(samples);
	Vecd		table2(samples);
	Vecd		table3(samples);
	Vecd		table4(samples);
	Vecd		table5(samples);
	Vecd		table6(samples);
	Vecd		table7(samples);

	// set options
	
	arg_format = arg_to_form(0,
		"", 										"usage: fc [options]",
		
		"%S", &filename,							"Specify scene file",
		
		"-allStats", ARG_FLAG(&allStats), 			"stats about all clusters",
		"-clus %S", &clusStr, 						"stats about one cluster",
		"-clusExt %S", &clusExtStr, 				"print cluster as obj",			
		"-showRest", ARG_FLAG(&showRest), 			"show rest of cluster too",			
		"-calcOBB %d", &calcOBB, 					"calc OBBs",
		"",						"    0 - use VOB",
		"",						"    1 - standard",
		"",						"    2 - convex hull",
		"",						"    3 - best fit axes using MER",
		"-table %d", &clusTable, 					"dump table as obj",
		"",						"    0 - actual (no-vis) PAs",
		"",						"    1 - lower bound",
		"",						"    2 - upper bound",
		"",						"    3 - table of ones (to get directional sphere)",
		"-volClus", ARG_FLAG(&volClus), 			"calc volume cluster",
		"-aa", ARG_FLAG(&setAA), 					"axis-aligned bboxes",
		"-metricStats", ARG_FLAG(&metricStats), 	"metric stats",
		0);

	// No arguments: print usage
	if (argc == 1)
	{
		cerr << "fc: collect face stats" << endl;
		arg_form_print(arg_format);
		exit(1);
	}
	// parse arguments
	if (arg_parse_argv(argc, argv, arg_format) < 0)
		exit(1);

	modelFile.SetPath(filename);
	StrConst exts[] = {"cmf", "mrb", "mmf", 0};
	if (modelFile.FindFileExtension(exts) < 0)
	{
		cerr << "couldn't find/bad extension: " << filename << endl;
		exit(-1);
	}
	modelFile.DecompressSetup();

	if (!model->Parse(modelFile.GetPath()))
	{
		cerr << "couldn't parse " << filename << endl;
		modelFile.DecompressCleanup();
		exit(-1);
	}
	else
	{
		cerr << "face clusters -- " << model->clusters.NumItems() << endl;
		cerr << "root clusters -- " << model->rootClusters.NumItems() << endl;
		cerr << "simplification vertices -- " << model->vertices.NumItems() << endl;
		modelFile.DecompressCleanup();
	}

	cerr << "sorting root clusters by size" << endl;
	SortRootClusters(model);
	cerr << "done." << endl;
	
	if (clusStr)
	{
		clus = ToClusterRef(model, clusStr);
		levels = FindDepth(model, clus);
		
		InitBox(box1);
		CalcAxisAlignedBoxPAs(model, clus, box1);
		CalcBoundingBoxPAs(model, clus, box2);

		FindDPs(model, clus, table1);
		FindActualPAs(model, clus, table2);
		FindPAsFromSumAreaNormal(model, clus, table3);
		FindPAsFromBoundingBox(model, clus, table4, box2);
		FindPAsFromBoundingBox(model, clus, table5, 0);
		FindPAsFromAxisAlignedBox(box1, table6);

		if (clusTable < 0)
		{
			cerr << "levels = " << levels << endl;
			cout << "# cos(d)  ActualPA(d)  SumAreaNormalPA(d)  BoxSampPA(d)  BoxAreaPA(d)  AABoxPA(d) totArea PA(n)" << endl;
			for (i = 0; i < samples; i++)
			{
				cout << table1[i] << ' '
					 << table2[i] << ' '
					 << table3[i] << ' '
					 << table4[i] << ' '
					 << table5[i] << ' '
					 << table6[i] << ' '
					 << model->Cluster(clus).totArea << ' '
					 << len(model->Cluster(clus).areaNormal) << ' '
					 << endl;

				Expect(table2[i] - table1[i] < 1e-6, "Bounds incorrect");
			}
		}
		else
		{
			if (clusTable == 0)
				WriteObjTable(table2);
			else if (clusTable == 1)
				WriteObjTable(table3);
			else if (clusTable == 2)
				WriteObjTable(table4);
			else
			{
				Vecd ct(samples, vl_1);
				WriteObjTable(ct);
			}
		}
	}
	else if (clusExtStr)
	{
		clusExt = ToClusterRef(model, clusExtStr);
		cout << "# cluster " << clusExtStr << " from " << filename << endl;
		PrintClusterAsObj(model, clusExt, showRest);
	}
	else if (allStats)
	{
		Double 	berr1, berr2, berr3, berr4, berr5, berr6;

		cout << "# cluster_ID level LB_MaxErr UB_MaxErr Tot_MaxErr LB_RMSErr UB_RMSErr Tot_RMSErr" << endl;

		for (i = model->faces.NumItems(); i < model->faces.NumItems() + model->clusters.NumItems(); i++)
		{
			Double 	area = model->Cluster(i).totArea;

			InitBox(box1);
			CalcBoundingBoxPAs(model, i, box1);

			FindActualPAs(model, i, table1);
			FindPAsFromSumAreaNormal(model, i, table2);
			FindPAsFromBoundingBox(model, i, table3, box1);

			berr1 = MaxError((table1 - table2));
			berr2 = MaxError((table1 - table3));
			berr3 = MaxError((table2 - table3));
			berr4 = RMSError((table1 - table2));
			berr5 = RMSError((table1 - table3));
			berr6 = RMSError((table2 - table3));

			levels = FindDepth(model, i);

			cout << i << ' '
				 << levels << ' '
				 << berr1 << ' '
				 << berr2 << ' '
				 << berr3 << ' '
				 << berr4 << ' '
				 << berr5 << ' '
				 << berr6 << ' '
				 << area
				 << endl;
		}
	}
	else if (calcOBB >= 0)
	{
		ProgramTimer	pt;

		pt.StartTimer();

		if (calcOBB == 0)
		{
			VolBounds vb;
			
			for (i = 0; i < model->rootClusters.NumItems(); i++)
			{
				vb.Clear();
				CalcBoundingBoxesVoB(model, model->rootClusters[i], vb);
			}
		}
		else if (calcOBB == 1)	
		{
			if (setAA)
			{
				VecTrans aa;

				aa[2] = vl_y;	// y up.
				aa[0] = vl_x;
				aa[1] = -Vector(vl_z);

				for (i = 0; i < model->clusters.NumItems(); i++)
					model->clusters[i].SetAxes(aa);
			}
							
			CalcBoundingBoxes(model);
			CalcBoundingBoxPAs(model);
		}
		else if (calcOBB == 2)
		{
			PointList	points;
			
			for (i = 0; i < model->rootClusters.NumItems(); i++)
				CH_CalcBoundingBoxes(model, model->rootClusters[i], points);
		}
		else if (calcOBB == 3)
		{
			ByteArray	marks(model->points->NumItems());
			IndexList	contour;

			for (i = 0; i < marks.NumItems(); i++)
				marks[i] = 0;
			
			for (i = 0; i < model->rootClusters.NumItems(); i++)
				FindBestFitAxes(model, model->rootClusters[i], contour, marks);

			CalcBoundingBoxes(model);
			CalcBoundingBoxPAs(model);
		}

		pt.StopTimer();
		
		cout << "calculated BBs in " << pt.GetTimer() << " seconds" << endl;
		model->WriteBinary();
	}
	else if (volClus)
	{
		ProgramTimer	pt;
		VolClus 		vc(model->faces, *model->points);
		FILE			*f;
		Int				id = 0;
		
		pt.StartTimer();
		vc.Create();
		pt.StopTimer();
		cout << "clus_time " << pt.GetTimer() << endl;
		//vc.Print(cout);
		vc.FindMetricStats();
		f = fopen("vol-clus.hier", "w");
		vc.DumpHier(f, id);
		fclose(f);
	}
	else if (metricStats)
	{
		FindVolStats(model);
	}
	
//	delete model; // XXX
		
	return(0);
}
