/*
	File:		VolBounds.cc

	Function:	

	Author:		Andrew Willmott

	Notes:		
*/

#include "VolBounds.h"
#include "VecUtil.h"
#include "Samples.h"
#include "String.h"

PointList VolBounds::tBoundsDir;

#define VB_AA
// axis-aligned samples
#define VB_BOUNDS_ONLY
// just find bounds with original directions; don't try to find
// new BB directions.
#define VB_SPHERE
// sample over sphere rather than random points

Void		GetSphereFaces(IndexList &triIndices);
PointList	MySpherePoints();
Int			ReverseSphereDir(Int i);

typedef NArray<VecTrans> TransformList;

static TransformList tFaces;
static IndexList tTriIndices;

Void CreateMaps()
{
	Int			i;
	VecTrans	trans;
	
	cerr << "initting face map" << endl;
		
	tFaces.Clear();

	for (i = 0; i < tTriIndices.NumItems(); )
	{
		col(trans, 0) = VolBounds::tBoundsDir[tTriIndices[i++]];
		col(trans, 1) = VolBounds::tBoundsDir[tTriIndices[i++]];
		col(trans, 2) = VolBounds::tBoundsDir[tTriIndices[i++]];

		tFaces.Append(inv(trans));
	}
}

VolBounds::VolBounds()
{
	Int		i;

	if (tBoundsDir.NumItems() == 0)
	{
		// init default direction set
#ifdef VB_AA
		// axis-aligned samples
		tBoundsDir.Append(Vector(vl_x));
		tBoundsDir.Append(Vector(vl_y));
		tBoundsDir.Append(Vector(vl_z));
		tBoundsDir.Append(-Vector(vl_x));
		tBoundsDir.Append(-Vector(vl_y));
		tBoundsDir.Append(-Vector(vl_z));
		tTriIndices = Indexes(
			0, 1, 2,
			5, 1, 0,
			3, 1, 5,
			2, 1, 3,

			2, 4, 0,
			0, 4, 5,
			5, 4, 3,
			3, 4, 2,

			IDX_END);
#else
#ifdef VB_SPHERE
		tBoundsDir = MySpherePoints();
		GetSphereFaces(tTriIndices);
#else		
		tBoundsDir.SetSize(128);
		for (i = 0; i < tBoundsDir.NumItems(); i++)
			tBoundsDir[i] = *((Point*) tSphereSamples[i]);		
#endif
#endif

		cerr << "VolBounds planes = " << tBoundsDir.NumItems() << endl;
		CreateMaps();
	}

	Clear();
}

Void VolBounds::Clear()
{
	Int		i;	

//	min.SetSize(tBoundsDir.NumItems());
	max.SetSize(tBoundsDir.NumItems());

	for (i = 0; i < max.Elts(); i++)
	{
		max[i] = -vl_inf;
//		min[i] = vl_inf;
	}
}

Void VolBounds::AddPoint(const Point &p)
{
	Vector 	dir;
	Double	b;
	Int		i;
	
	for (i = 0; i < max.Elts(); i++)
	{
		dir = tBoundsDir[i];

		b = dot(dir, p);
		if (b > max[i])
			max[i] = b;
//		else if (b < min[i])
//			min[i] = b;
	}
}

#ifdef VB_TEST3
Double VolBounds::FindBound(const Vector &d)
{
	Double result = 0.0;
	
	if (d[0] > 0.0)
		result += d[0] * max[0];
	else
		result -= d[0] * max[3];

	if (d[1] > 0.0)
		result += d[1] * max[1];
	else
		result -= d[1] * max[4];

	if (d[2] > 0.0)
		result += d[2] * max[2];
	else
		result -= d[2] * max[5];

	return(result);
}
#else
Double VolBounds::FindBound(const Vector &d)
{
	Vector	test;
	Int		i;
		
	// NOTE: can't just find closest intersection with bounding
	// plane, because that may be too close -- we're finding tightest
	// bounding plane, not closest point. So must use stu formulation.

	// find the three vectors containing d

	for (i = 0; i < tFaces.NumItems(); i++)
	{
//		cout << "check " << tFaces[i];
		test = tFaces[i] * d;
//		cout << "coords: " << test << endl << endl;
		if (test[0] >= 0.0 && test[1] >= 0.0 && test[2] >= 0.0)
		{
//			cout << "Got it" << endl;
			return(
				test[0] * max[tTriIndices[3 * i + 0]] +
				test[1] * max[tTriIndices[3 * i + 1]] +
				test[2] * max[tTriIndices[3 * i + 2]]
			);
		}
	}

	_Error("incomplete directional sampling");

	return(0.0);
}
#endif

#ifdef BROKEN
Double VolBounds::FindMinBoundInPlane(const Vector &n, Vector &d)
// what is the smallest extent of the object in the plane defined by n,
// and in what direction?
{
	Double	dp, t, result = vl_inf, oldDP;
	Vector	resultDir;
	Int		i, oldIdx;

	// consider a plane through dir[i] * bounds[i]
	// consider the line where it intersects plane n
	// what is the min distance from this line to origin?

	// lies in projection of dir onto d

	for (i = 0; i < bounds.Elts(); i++)
	{
		// if theta = angle between dir[i] and n
		// sin(theta) = bounds / bn, where bn is the bound in the plane
		// bn = bounds / sqrt(1 - dp^2)

		dp = dot(n, tBoundsDir[i]);
		t = sqrt(1.0 - sqr(dp));
		if (t > 0.0)
		{
			t = bounds[i] / t;
			// find the minimum of all the bounds
			if (t < result)
			{
				result = t;
				resultDir = norm(tBoundsDir[i] - dp * n);
				// norm should be okay 'cause t > 0
				oldDP = dp;
				oldIdx = i;
			}
		}
	}

	dp = dot(resultDir, n);
	Assert(dp < 1e5, "bad dir");
	d = resultDir;

	return(result);
}
#endif

Double VolBounds::FindMinBoundInPlane(const Vector &n, Vector &d)
{
	Vector	u, v, dir, minDir;
	Double bound, minBound, theta, i;
		
	u = norm(FindOrthoVector(n));
	v = cross(n, u);

	for (i = 0.0; i < 2.0; i += 0.05)
	{
		theta = vl_pi * i;
		dir = cos(theta) * u + sin(theta) * v;
		bound = FindBound(dir);
		if (bound < minBound)
		{
			minBound = bound;
			minDir = dir;
		}
	}

	d = minDir;

	return(minBound);
}

Void VolBounds::MergeBounds(VolBounds &b2)
{
	Int		i;

	for (i = 0; i < max.Elts(); i++)
	{
		max[i] = Max(max[i], b2.max[i]);
//		min[i] = Min(min[i], b2.min[i]);
	}
}

#ifndef VB_BOUNDS_ONLY
Void VolBounds::FindBoundingBox(VecTrans &axis, Point &min, Point &max)
{
	Double	minWidth = vl_inf, thickness;
	Int		minIdx = 0, i;
	Int		minRevIdx, revIdx;
	
	for (i = 0; i < max.Elts() / 2; i++)
	{
//		thickness = max[i] - min[i];
		revIdx = ReverseSphereDir(i);
		thickness = max[i] + max[revIdx];

		if (thickness < minWidth)
		{
			minWidth = thickness;
			minIdx = i;
			minRevIdx = revIdx;
		}
	}
	
	axis[0] = tBoundsDir[minIdx];
	max[0] = max[minIdx];
	min[0] = -max[minRevIdx];

#ifdef UNFINISHED
	for (i = 0; i < max.Elts(); i++)
	{
		c = dot(tBoundsDir[i], axis[0]);
		s = sqrt(1.0 - sqr(c));
		if (s > 0.3)
		{
			thickness = (max[i] - min[i]) / s;
			if (thickness < minWidth)
			{
				minWidth = thickness;
				minIdx = i;
			}
		}
	}

	c = dot(tBoundsDir[i], axis[0]);
	s = sqrt(1.0 - sqr(c));
#endif
	
	
	axis[2] = cross(axis[0], axis[1]);
	max[2] = FindBound(axis[2]);
	min[2] = -FindBound(-axis[2]);
}
#else
Void VolBounds::FindBoundingBox(VecTrans &axis, Point &min, Point &max)
{
//	axis[0] = vl_x;	
//	axis[1] = vl_y;	
//	axis[2] = vl_z;	

	max[0] = FindBound(axis[0]);
	max[1] = FindBound(axis[1]);
	max[2] = FindBound(axis[2]);
	min[0] = -FindBound(-axis[0]);
	min[1] = -FindBound(-axis[1]);
	min[2] = -FindBound(-axis[2]);
}
#endif
