/*
	File:		VolClus.cc
	
	Purpose:	Implements volume clustering
	
	Author:		Andrew Willmott
	
*/

#include <stdio.h>
#include "VolClus.h"
#include "VecUtil.h"
#include "Draw.h"
//#include "Samples.h"

const Int kMinClusterElems = 32;
const Int kMaxClusterLevels = 16;

#define DBG_COUT if (0) cerr

VolClus::VolClus(FaceIdxArray &f, PointList &p) : elems(), faces(f), points(p)
{
	Int i;
	
	centre = vl_0;
	
	for (i = 0; i < 8; i++)
		child[i] = 0;		

}

VolClus::~VolClus()
{
	Int i;

	DBG_COUT << "deleting cluster" << endl;

	for (i = 0; i < 8; i++)
		delete child[i];
}

Void VolClus::CreateBounds()
{
	Int		i;
	
	min = max = ElemCentre(0);
	
	for (i = 0; i < elems.NumItems(); i++)
		ElemUpdateBounds(elems[i], min, max);

	centre = (min + max) / 2.0;
}

Point VolClus::ElemCentre(Int i)
{
	Point centre;
	
	centre = points[faces[i].v[0]];
	centre += points[faces[i].v[1]];
	centre += points[faces[i].v[2]];

	return(centre / 3.0);
}

Void VolClus::ElemUpdateBounds(Int i, Point &min, Point &max)
{
	UpdateBounds(points[faces[i].v[0]], min, max);
	UpdateBounds(points[faces[i].v[1]], min, max);
	UpdateBounds(points[faces[i].v[2]], min, max);
}

GCLReal VolClus::ElemArea(Int i)
{
	Vector areaNorm;
	
	CalcTriAreaNormal(
			points[faces[i].v[0]],
			points[faces[i].v[1]],
			points[faces[i].v[2]],
			areaNorm
	);

	return(len(areaNorm) / 2.0);
}

Void VolClus::Create()
{
	Int		i;
	
	elems.SetSize(faces.NumItems());

	for (i = 0; i < elems.NumItems(); i++)
		elems[i] = i;

	CreateBounds();

	clusterVol = BoxVol(min, max);
	level = 0;
	CreateHierarchy();
}

Void VolClus::PushElems()
{
	Vector		d1, d2;
	Int			i, j;
	Point		emin, emax, eCentroid;
	IndexList	newElems;

	for (j = 0; j < elems.NumItems(); j++)
	{
		emin = emax = ElemCentre(elems[j]);

		ElemUpdateBounds(elems[j], emin, emax);
		d1 = emax - emin;
		d2 = (max - min) / 2.0;
		
//		DBG_COUT << "size check: " << d1 << " < " << d2 << endl;

		// if the element is bigger in size than an octant, add it to this
		// level of the cluster hierarchy.
		if (d1[0] > d2[0] || d1[1] > d2[1] || d1[2] > d2[2])
		{
			newElems.Append(elems[j]);
			continue;
		}
			
		// find the octant the element's centroid belongs to...
		eCentroid = (emin + emax) / 2.0;
		d1 = eCentroid - centre;
		if (sqrlen(d1) < 1e-8)
		{
			// if it's close to the centre, add it to this level after all.
			newElems.Append(elems[j]);
			continue;
		}
		
		i = 0;
		if (d1[0] >= 0.0) i += 1;
		if (d1[1] >= 0.0) i += 2;
		if (d1[2] >= 0.0) i += 4;
		
		// add it to the given octant
		if (child[i] == 0)
			child[i] = new VolClus(faces, points);
		child[i]->elems.Append(elems[j]);
	}
	
	elems.Replace(newElems);
}
 
VolClus *VolClus::CreateHierarchy()
{
	Int			i, n = 0;
	VolClus		*lastChild = 0;	

	clusterArea = 0.0;
	clusterVol = BoxVol(min, max);

	if (elems.NumItems() >= kMinClusterElems && level < kMaxClusterLevels)
	{
		PushElems();

		for (i = 0; i < 8; i++) 
			if (child[i]) 
			{	
				lastChild = child[i];
				child[i]->CreateBounds();
				child[i]->level = level + 1;
				child[i] = child[i]->CreateHierarchy();
				clusterArea += child[i]->clusterArea;
				n++;
			}
	}
	
	// if there are no elements at this level, and we only have one child,
	// we can replace this node with its child.
	if (elems.NumItems() == 0 && n <= 1)
	{
		DBG_COUT << "collapsing cluster node" << endl;
		return(lastChild);
	}
	else if (n <= 1)
		DBG_COUT << "didn't collapse cluster node: has kids = " << elems.NumItems() << endl;

	for (i = 0; i < elems.NumItems(); i++)
		clusterArea += ElemArea(elems[i]);
		
	return(this);
}

Void VolClus::Print(ostream &s)
{
	Int		i;
	
	s << "level " << level << " cluster: " << endl;
	s << elems.NumItems() << " patch children" << endl;
	s << "vol " << clusterVol << " area " << clusterArea << endl;
	for (i = 0; i < 8; i++) 
		if (child[i]) 
			child[i]->Print(s);
}

#ifdef OLD

Void VolClus::DrawNodeElem(Renderer &r)
{
	Int		i;

	DrawLeafElem(r);
	for (i = 0; i < 8; i++)
		if (child[i])
			child[i]->DrawElem(r);
}

Void VolClus::DrawLeafElem(Renderer &r)
{
#ifdef RAD_VIS
	if (eltHilite != 0)
	{
		if (eltHilite == 1)
			r.C(Colour4(cRed, 0.3));
		else if (eltHilite == 2)
			r.C(Colour4(cYellow, 0.3));
		else if (eltHilite == 3)
			r.C(Colour4(cGreen, 0.3));
		PaintBox(r, min, max);
	}
	else if (gRadControl->outlineClusters)
	{
		r.C(HSVCol(hsvYellow + level * 60, 1, 1));
		FrameBox(r, min, max);
		
		DrawPatches(r);
	}
#endif
}

Void VolClus::DrawPatches(Renderer &r)
{
	Int	i;

	for (i = 0; i < elems.NumItems(); i++)
		elems[i]->DrawLeafElem(r);

	for (i = 0; i < 8; i++)
		if (child[i])
			child[i]->DrawPatches(r);
}

#endif


Int VolClus::DumpHier(FILE *f, Int &id)
/// Dump the cluster hierarchy as a .hier file.
{
	Int			i;
	IndexList	c;

	for (i = 0; i < 8; i++)
		if (child[i])
			c.Append(child[i]->DumpHier(f, id));

	fprintf(f, "e %d", id);
	for (i = 0; i < c.NumItems(); i++)
		fprintf(f, " %d", c[i]);
	fprintf(f, "\n");
	
	return(id++);
}

Void VolClus::FindMetricStats()
{
	Int		n;
	Double	sumVol;
	
	n = 0;
	sumVol = 0.0;
	CalcMetricStats(n, sumVol);

	cout << "avg_clus_vol " << sumVol / n << endl;
}

Void VolClus::CalcMetricStats(Int &n, Double &sumVol)
{
	for (Int i = 0; i < 8; i++)
		if (child[i])
			child[i]->CalcMetricStats(n, sumVol);

	n++;
	sumVol += clusterVol;
}
