/*
	File:			Camera.cc

	Function:		See header file

	Author(s):		Andrew Willmott

	Copyright:		Copyright (c) 1995-1997, Andrew Willmott
*/

#include "gcl/Camera.h"


Camera::Camera() : 
	position(0, 0, 4),
	fov(30.0),
	aspect(1.0),
	zoom(1.0),
	clipNear(0.01),
	clipFar(100.0),
	up(vl_y),
	dir(Vector(vl_z) * -4.0),
	lhand(0)
{
	sceneOrient = vl_I;
	sceneOffset = vl_I;
}

Void Camera::SetClipping(GCLReal nearPlane, GCLReal farPlane)
{
	clipNear = nearPlane;
	clipFar = farPlane;
}

Void Camera::SetFOV(GCLReal fieldOfView)
{
	fov = fieldOfView;
}

Void Camera::SetAspect(GCLReal aspectRatio)
{
	aspect = aspectRatio;
}

Void Camera::SetPosition(Point pos)
{
	Point at = position + dir;
	position = pos;
	dir = at - position;
}

Void Camera::SetLookat(Point at)
{
	dir = at - position;
}

Void Camera::SetViewUp(Vector vup)
{
	up = norm(vup);
}

Void Camera::SetZoom(GCLReal zoomFactor)
{
	if (zoomFactor > 1e-6)
		zoom = zoomFactor;
	else
		zoom = 1e-6;
}

const GCLReal kProjEpsilon = 0.01;

Transform Camera::ModelMatrix() const
{
	Vector		xAxis, yAxis, zAxis;
	Transform	result, align;

	zAxis = -norm(dir);
	xAxis = cross(up, zAxis);
	
	if (len(xAxis) < kProjEpsilon)
		xAxis = norm(cross(vl_z, zAxis));
	else
		xAxis = norm(xAxis);
	
	yAxis = norm(cross(zAxis, xAxis));
		
	// Create a matrix which will align scene with the xAxis, yAxis, zAxis
	align[0] = MVector4(xAxis[0], xAxis[1], xAxis[2], 0.0);
	align[1] = MVector4(yAxis[0], yAxis[1], yAxis[2], 0.0);
	align[2] = MVector4(zAxis[0], zAxis[1], zAxis[2], 0.0);
	align[3] = vl_w;
#ifndef VL_ROW_ORIENT
	align = trans(align);
#endif
	
	// first: translate eyepoint to origin, rotate so camera lies on +ve z axis,
	// and apply rotation and offset around lookat. We apply rotations first,
	// because we want translations to be specified in the camera's coordinate
	// system. 
	// second: translate camera position to origin.
	
	result = sceneOffset * sceneOrient * align * Shift(-(position + dir));
	result = Shift(Vector(0, 0, -len(dir))) * result;

	return(result);
}

Transform Camera::ProjMatrix() const
{
	Transform	project(vl_I);
		

	if (lhand)
	{
		project[3][3] = 0;
		project[2][2] = -(clipFar + clipNear) / (clipFar - clipNear);
#ifndef VL_ROW_ORIENT
		project[2][3] = - 2 * clipFar * clipNear / (clipFar - clipNear);
		project[3][2] = -1;
#else
		project[3][2] = - 2 * clipFar * clipNear / (clipFar - clipNear);
		project[2][3] = -1;
#endif
			
		project[0][0] = 1.0 / tan(DegsToRads(fov / zoom));
		project[1][1] = aspect * project[0][0];
	}
	else
	{
		project[3][3] = 0;
		project[2][2] = -(clipFar + clipNear) / (clipFar - clipNear);
#ifndef VL_ROW_ORIENT
		project[2][3] = - 2 * clipFar * clipNear / (clipFar - clipNear);
		project[3][2] = -1;
#else
		project[3][2] = - 2 * clipFar * clipNear / (clipFar - clipNear);
		project[2][3] = -1;
#endif
			
		project[0][0] = 1.0 / tan(DegsToRads(fov / zoom));
		project[1][1] = aspect * project[0][0];
	}

	return(project);
}

Void Camera::SetSceneOrient(const Transform &t)
{
	sceneOrient = t;
}

Void Camera::SetSceneOffset(const Transform &t)
{
	sceneOffset = t;
}

ostream &operator << (ostream &s, Camera &c)
{
	s << c.ProjMatrix() << endl << c.ModelMatrix();
	
	return(s);
}

// --- Camera utilities -------------------------------------------------------

Void FindWorldRay(
			const Coord		&c,
			const Transform &inverseW2S,
			Point			&from,
			Vector			&dir
		)
{
 	HPoint eye(c[0], c[1], 0.0, 1.0);
	HPoint target(c[0], c[1], 1.0, 1.0);	

	eye = inverseW2S * eye;
	target = inverseW2S * target;
	
	from = proj(eye);
	dir = proj(target);
	dir -= from;
}


// --- OrthoCamera ------------------------------------------------------------

Transform OrthoCamera::ProjMatrix() const
{
	return(Scale(Vector(zoom, zoom, 1)));
}
