/*
	File:			Camera.h
	
	Function:		Defines a camera. When sent to a Renderer, this defines
					the current view (and perspective) transformations.
					A camera can be embedded in a scene: see SceneObjects.cc

	Author(s):		Andrew Willmott

	Copyright:		Copyright (c) 1995-1999, Andrew Willmott
 */
 
#ifndef __Camera__
#define __Camera__

#include "gcl/Geometry.h"
#include <iostream.h>

class Camera
// Maps view into viewport [-1,1] x [-1,1]
{
public:	
	Camera();
	
	Void		SetClipping(GCLReal nearPlane, GCLReal farPlane);
				// Set standard camera parameters.
	Void		SetFOV(GCLReal fieldOfView);
	Void		SetAspect(GCLReal aspect);
	Void		SetPosition(Point pos);
	Void		SetLookat(Point at);	
	Void		SetViewUp(Vector at);	
	
	Void		SetZoom(GCLReal zoomFactor);
				// Scale the scene around the lookat point.
	Void		SetSceneOrient(const Transform &t);
				// Rotate the scene around the lookat point.
	Void		SetSceneOffset(const Transform &t);
				// Offset the lookat point from its original location.
		
	virtual Transform ProjMatrix() const;
				// Return the projection matrix from Camera space to
				// Display Space ([-1,1] ^ 2)
	Transform 	ModelMatrix() const;
				// Return the transformation from World space to Camera
				// space
	
protected:
	
	Point		position;
	Point		up;					// up vector, normalised
	Point		dir;				// vector from camera position to lookat point.
	GCLReal		clipNear, clipFar;	// near and far clipping planes
	GCLReal		fov;				// field of view in degrees
	GCLReal		aspect;				// aspect ratio
	Int			lhand;				// left handed coordinate system?
	Int 		ortho;				// orthogonal camera 
									// XXX split out into own class
	
	GCLReal		zoom;
	Transform	sceneOrient;
	Transform	sceneOffset;
};

class OrthoCamera : public Camera
{
	Transform 	ProjMatrix() const;
};

ostream &operator << (ostream &s, Camera &c);

/*
	FindWorldRay maps from display coords ([-1, 1], [-1, 1]) to
	a ray through the world. We assume our world-to-screen matrix
	maps from some form of view pyramid to the box from -1 to 1 in x and y
	and 0 to 1 in z. 
		
	Note that this works regardless of whether a perspective or orthogonal
	projection is used; the only thing you need is the WS matrix.
*/

Void FindWorldRay(
			const Coord		&c,
			const Transform &inverseW2S,
			Point			&from,
			Vector			&dir
		);

#endif
