/*
    File:       ColourSystem.h

	Function:	Provides a class to help with colour space conversions to
				and from three-primary display devices.

    Author:     Andrew Willmott

    Copyright:  (c) 1999-2000, Andrew Willmott
*/

#ifndef __ColourSystem__
#define __ColourSystem__

#include "Colour.h"
#include "Geometry.h"
#include "VecUtil.h"

typedef Vec2f Chroma;

class ColourSystem
{
public:
	ColourSystem(
			Chroma 		primary1,
			Chroma 		primary2,
			Chroma 		primary3,
			Chroma 		whitePoint,
			ClrReal		whiteLum = 1.0
		);

	// transforming to and from CIE XYZ colour space.

	Colour		GamutToXYZ(const Colour &csClr)
				///  Convert device colour to CIE XYZ colour.
				{ return(toXYZ * csClr); };
	Bool		XYZIsOutOfGamut(const Colour &cieXYZ);
				/**< Returns true if the colour can't be represented by
					 this device. */

	Colour 		XYZToGamut(const Colour &cieXYZ);
				/**< Convert CIE XYZ colour to device colour, and
					 constrain the result to be within its gamut. */
	Colour		XYZToGamutClip(const Colour &cieXYZ);
				///< Do the same, but also clip result to [1, 1, 1]

	Bool		ChromaIsOutOfGamut(const Chroma &cieChroma);
				/**< Returns true if the colour can't be represented by 
					 this device. */
	Colour 		ChromaToGamut(const Chroma &cieChroma);
				/**< Convert CIE chromaticity to colour system colour of
					 unit luminance. */


	ClrReal		Luminance(const Colour &csClr)
				///  Finds the luminance of csClr.
				{ return(dot(toXYZ[1], csClr)); };


	// colour conversion matrices
	ClrMat		toXYZ;
	ClrMat		fromXYZ;

	// chromaticities of the three primaries & the chroma/luminance of the
	// white point
	Chroma		prim0; 
	Chroma		prim1; 
	Chroma		prim2; 
	Chroma		white;
	ClrReal		whiteLum;;
};

// current colour system for display output. defaults to csEBU, but
// might be necessary to set it to something more appropriate.

extern ColourSystem &csDisplay;

// chromaticities of standard illuminants
extern const Chroma csIllumA;		// Incandescent light
extern const Chroma csIllumB;		// Solar light
extern const Chroma csIllumC;		// For NTSC television
extern const Chroma csIllumD55;		// Photography
extern const Chroma csIllumD65;		// Overcast Daylight
extern const Chroma csIllumD75;
extern const Chroma csIllumE;		// Uniform illuminant

// luminances of standard illuminants, in candela/watt

extern const ClrReal csMaxLum;
extern const ClrReal csWhiteLum;
extern const ClrReal csD65Lum;
extern const ClrReal csALum;
extern const ClrReal csBLum;


// Colour systems!
extern ColourSystem csNTSC;
extern ColourSystem csPAL_SECAM;
extern ColourSystem csEBU;		// same as CCIR709
extern ColourSystem csSMPTE;
extern ColourSystem csShortPersistencePhosphors;
extern ColourSystem csLongPersistencePhosphors;
extern ColourSystem csDellPhosphors;

#endif
