/*
	File:		Hermite.cc

	Function:	

	Author:		Andrew Willmott

	Notes:		
*/

#include "gcl/Hermite.h"


/* This basis takes [x(0),x(1),x'(0),x'(1)] as knot points   
   Each column holds the coefficients of one of the Hermite 
   polynomials. */

static Transform gHermiteMatrix(
	 1.0,  0.0,  0.0,  0.0, 
	 0.0,  0.0,  1.0,  0.0,
	-3.0,  3.0, -2.0, -1.0,
	 2.0, -2.0,  1.0,  1.0
);


/*
	Calculate hermite coefficients.

	t0, y0, ySlope0 are start time, value, and slope.
	t1, y1, ySlope1 are end time, value, and slope.
	t1 had better not equal t0.
		
	For automatic slope control, set 
		ySlope1 = (y2 - y0) / (t2 - t0);
		ySlope0 = (y1 - y-1) / (t1 - t-1);
*/

Void CalcHermiteCoeffs(
		GCLReal		t0,
		GCLReal		t1,
		GCLReal		y0,
		GCLReal		y1,
		GCLReal		ySlope0,
		GCLReal		ySlope1,
		Cubic		&result
	)
{
	result = gHermiteMatrix * 
		Cubic(y0, y1, ySlope0 * (t1 - t0), ySlope1 * (t1 - t0));
}

GCLReal EvalCubic(Cubic &cubicCoeffs, GCLReal t, GCLReal t0, GCLReal t1)
//	evaluate y(t), using coeffs computed by CalcHermiteCoeffs().
//	t0, t1 have to be same as you supplied to CalcHermiteCoeffs().
//	t should be between t0 and t1, inclusive.
{
	GCLReal	u = (t - t0) / (t1 - t0);
	Cubic	uCubic(1, u, sqr(u), sqr(u) * u);

	return(dot(uCubic, cubicCoeffs));
}


/*
	Here's a dumb sample program that interpolates values of y between
	frame 0 and frame 30:
*/

#ifdef HM_DEBUG
#include <stdio.h>

main()
{
	Cubic	coeffs;
	GCLReal	t0 = 0, t1 = 30;
	GCLReal	y0 = 1,  y1 = 3;
	GCLReal	dy0 = .2, dy1 = 0;
	Int		t;

	CalcHermiteCoeffs(t0, t1, y0, y1, dy0, dy1, coeffs);

	for(t = 0; t < 31; t++)
		printf("\ny(%d) = %f", t, EvalCubic(coeffs, t, t0, t1) );
}

#endif
