/*
	File:		ImgView.cc

	Function:	Simple Image Viewer

	Author:		Andrew Willmott

	Notes:		
*/

#define GCL_WATCH_FILE

#include "gcl/XGraphicsSystem.h"
#include "cl/Timer.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "arg.h"

enum IPCmd 
{
	ipNone,
	ipReload,
	ipNext,
	ipPrev,
	ipQuit
};

class ImagePane : public XBackedPane
{
public:
	ImagePane() : XBackedPane(), command(ipNone), offscreen(0), closed(false)
	{};

	Void			HandleKey(Char c);
	Void			PrintKeys();
	Void			PrintInfo();
	Void			CheckFile();
	Int				Load(StrConst filename);
	
	IPCmd			command;
	FileName 		imageFile;
	RGBAImage		image;
	RGBEImage		radImage;
	Long			timeStamp;
	Int				offWidth, offHeight;
	Bool			closed;
	Bool			loaded;
	Double			alpha;
	XOffscreenPane	*offscreen;

	static XGraphicsSystem	*sXgs;
};

XGraphicsSystem *ImagePane::sXgs = 0;

typedef ImagePane *ImagePanePtr;

// XXX image stuff -- quick hack

#include "VecUtil.h"

struct ImageStats
{
	Double	avgLum;
	Double	avgLogLum;
	Double	minLum;
	Double	maxLum;
	Colour	avgClr;
	Colour	minClr;
	Colour	maxClr;
	Int		numSamples;
};

Void FindImageStats(Image &img, ImageStats &stats)
{		
	Int		x, y;
	Colour	c;
	Double	lum;
	
	stats.numSamples = 0;
	stats.avgLogLum = 0.0;
	stats.avgLum = 0.0;
	stats.minLum = vl_inf;
	stats.maxLum = -vl_inf;
	stats.avgClr.MakeZero();
	stats.minClr.MakeBlock(vl_inf);
	stats.maxClr.MakeBlock(-vl_inf);
	
	for (y = 0; y < img.Height(); y++)
		for (x = 0; x < img.Width(); x++)
		{
			c = img.GetPixel(x, y);
			lum = dot(cRGBToLum, c);

			if (lum > 0.0)
			{
				stats.avgClr += c;
				stats.avgLum += lum;
				stats.avgLogLum += log(lum);
				FindMinCmpts(c, stats.minClr, stats.minClr);
				FindMaxCmpts(c, stats.maxClr, stats.maxClr);
				stats.minLum = Min(lum, stats.minLum);
				stats.maxLum = Max(lum, stats.maxLum);
				stats.numSamples++;
			}
		}

	stats.avgLum /= stats.numSamples;
	stats.avgLogLum /= stats.numSamples;
	stats.avgClr /= stats.numSamples;
}

Void LinearMapImage(const Image &img, Double start, Double slope, Image &imgOut)
{		
	Int		x, y;
	Colour	c, cStart;

	imgOut.SetSize(img.Width(), img.Height());
	cStart.MakeBlock(start);
	
	for (y = 0; y < img.Height(); y++)
		for (x = 0; x < img.Width(); x++)
		{
			c = img.GetPixel(x, y);
			c -= cStart;
			c *= slope;
			imgOut.SetPixel(x, y, ClipColour(c));
		}
}

Void DoToneMap(Image &radImage, Image &image)
{
	ImageStats	stats;
	Double		scale, EV;
	
	FindImageStats(radImage, stats);

	// roughly approximate exposure value for ISO 100 film
	EV = stats.avgLogLum / log(2) - 1;
	// crop arbitrarily to "sensible" values
	EV = Clip(EV, 6.0, 17.0);
	// use this to scale irradiances so average level is 0.18,
	// a la most light meters
	scale = 0.18 * exp(-(EV + 1) * log(2));
	LinearMapImage(radImage, 0.0, scale, image);
}

// ----

Int ImagePane::Load(StrConst filename)
{
	Int		err;
	Bool	justCreated = false;
	
	loaded = false;
	imageFile.SetPath(filename);

	if (imageFile.GetExtension() == "pic")
	{
		err = radImage.LoadPIC(imageFile.GetPath());
		if (!err)
			DoToneMap(radImage, image);
	}
	else
		err = image.Load(imageFile);

	if (err != 0)
	{
		cerr << "image read failed." << endl;
		return(err);
	}

	if (!xgs)
	{
		sXgs->CreateWindow(this, imageFile.GetFile(), image.Width(), image.Height());
		justCreated = true;
	}
	
	if (!offscreen || image.Width() > offWidth || image.Height() > offHeight)
	{
		if (offscreen)
			delete offscreen;

		offscreen = new XOffscreenPane;
		xgs->CreateOffscreen(offscreen, image.Width(), image.Height());
		offWidth = image.Width();
		offHeight = image.Height();
		Attach(offscreen);
	}
	offscreen->PutImage(image);

	if (!justCreated)
	{
		SetTitle(imageFile.GetFile());
		Resize(image.Width(), image.Height());
	}
	HandleExpose();

	timeStamp = imageFile.GetTimeStamp();
	command = ipNone;
	loaded = true;

	return(0);
}

Void ImagePane::CheckFile()
{
	if (loaded && imageFile.GetTimeStamp() > timeStamp)
		Load(imageFile.GetPath());
}

Void ImagePane::HandleKey(Char c)
{
	switch (c)
	{
	case 'h':
	case '?':
		PrintKeys();
		break;
	case 'i':
		PrintInfo();
		break;
	case 'd':
		{
			FileName outFile;
			outFile = imageFile;
			outFile.SetExtension("tif");
			image.Save(outFile);
		}
		break;
	case 'r':
		Load(imageFile.GetPath());
		break;
	case '1':
		image.Transform(RGBSaturate(alpha));	
		offscreen->PutImage(image);
		HandleExpose();
		break;
	case '2':
		image.Transform(RGBSaturate(2.0 - alpha));
		offscreen->PutImage(image);
		HandleExpose();
		break;
	case '3':
		image.Transform(RGBHueRotate(alpha));
		offscreen->PutImage(image);
		HandleExpose();
		break;
	case '4':
		image.Clear(cRed);
		offscreen->PutImage(image);
		HandleExpose();
		break;
	case ',':
	case 0x1D:
		command = ipPrev;
		xgs->SignalDone();
		break;
	case ' ':
	case '.':
	case 0x1C:
		command = ipNext;
		xgs->SignalDone();
		break;
	case 'w':
		Hide();
		closed = true;
		break;
	case 'q':
	case 0x1B:
		command = ipQuit;
		xgs->SignalDone();
		break;
									
	default:
		XBackedPane::HandleKey(c);
	}
}

Void ImagePane::PrintKeys()
{
	cout << "Viewer control:" << endl
		 << "  r - reload image" << endl
		 << "  d - dump tiff image" << endl
		 << "  . or <space> - move on to next image" << endl
		 << "  , - move back to previous image" << endl
		 << "  i - print image information" << endl
		 << "  q or <esc> - quit" << endl
		 << "  w - close window" << endl
		 << endl;
}

Void ImagePane::PrintInfo()
{
	cout << String().Printf("Image %s is %d x %d", imageFile.GetPath().CString(),
		image.Width(), image.Height()) << endl; 
}

class ImgViewApp
{
public:
	ImgViewApp();
	
	Void				SetOptions(Int argc, Char **argv);
	Int					Run();

// options
	static Void			GetFileArgs(Int argc, Char *argv[]);

	static Char			**files;
	static Int			numFiles;
	Int					noFork;
	Int					all;

	Double				alpha;
	Int					doSat;
};

Char	**ImgViewApp::files = 0;
Int		ImgViewApp::numFiles = 0;

ImgViewApp::ImgViewApp()
{
}

Void ImgViewApp::GetFileArgs(Int argc, Char *argv[])
{
	files = argv;
	numFiles = argc;
}

Void ImgViewApp::SetOptions(int argc, char **argv)
{
	arg_form	*arg_format;
	Int			formats;
	
	alpha = 1.0;
	
	arg_format = arg_to_form(0,
		"", 
			"Usage: imgview file1 file2 ... [options]",
		"", ARG_SUBR(GetFileArgs), 
			"",
		"-noFork", ARG_FLAG(&noFork),
			"Don't fork into background.",
		"-all", ARG_FLAG(&all),
			"Open all images at once.",
		"-alpha %F", &alpha,
			"Set alpha for image operation",
		"-sat", ARG_FLAG(&doSat),
			"Change saturation of the image",
		"-formats", ARG_FLAG(&formats),	
			"List supported file formats",
		0);

	if (argc == 1)
	{
		arg_form_print(arg_format);
		cout << endl;
		exit(0);
	}

	if (arg_parse_argv(argc, argv, arg_format) < 0)
		exit(1);

#ifdef DEBUG
	noFork = true;
#endif

	if (formats)
	{
		Image::PrintSupportedFormats(cout);
		exit(0);
	}
}

Int ImgViewApp::Run()
{
	Int		i;

	if (numFiles < 1)
		return(1);

	if (!noFork && fork())
		return(0);

	ImagePane::sXgs = new XGraphicsSystem;

	if (all)
	{
		ImagePanePtr 	*panes = new ImagePanePtr[numFiles];
		Int				closed = 0;
		
		for (i = 0; i < numFiles; i++)
		{
			panes[i] = new ImagePane;
			panes[i]->alpha = alpha;
			panes[i]->Load(files[i]);
		}

		while (closed < numFiles)
		{
			if (ImagePane::sXgs->RunFor(1.0))
				break;
			for (i = 0; i < numFiles; i++)
				if (panes[i])
				{
					if (panes[i]->closed)
					{
						delete panes[i];
						panes[i] = 0;
						closed++;
					}
					else
						panes[i]->CheckFile();
				}
		}
		
		for (i = 0; i < numFiles; i++)
			delete panes[i];
	}
	else
	{
		Int 		curImage = 0;
		ImagePane	*imagePane = new ImagePane;

		imagePane->Load(files[0]);
		imagePane->alpha = alpha;
		
		do
		{
			ImagePane::sXgs->RunFor(1.0);

			imagePane->CheckFile();
			
			switch (imagePane->command)
			{
			case ipPrev:
				if (curImage > 0)
					curImage--;
				else
					curImage = numFiles - 1;
				imagePane->Load(files[curImage]);
				break;
			case ipNext:
				curImage++;
				if (curImage == numFiles)
					curImage = 0;
				imagePane->Load(files[curImage]);
				break;
			}
		}
		while (imagePane->command != ipQuit);

		delete imagePane;
	}
	
	return(0);
}

int main(int argc, char **argv)
{
	ImgViewApp	app;
	
	app.SetOptions(argc, argv);
		
	return(app.Run());
}
