/*
	File:		MRModel.cc

	Function:	Implements a multiresolution polygonal model

	Author:		Andrew Willmott

	Notes:		
*/

#include "gcl/MRModel.h"
#include "cl/String.h"
#include "gcl/VecUtil.h"
#include "gcl/Draw.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#ifdef DEBUG
#define MRV_COUT if (true) cerr
#else
#define MRV_COUT if (false) cerr
#endif

#define MR_MEM_MAP

#ifdef MR_MEM_MAP
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#endif

#include <new.h>


/*! \class FaceClusList

	Note: clusters are numbered 0 (first leaf) to 2f - 1 (root)
	where f = number of faces in the model.
	clusters corresponding to the leaf faces are not stored, to
	save on memory.
	the cluster ID corresponds to the cluster position in the
	clustersActive list.
	however, this is *not* the case for the clusters themselves.
	they are stored in reverse order, so we can easily truncate the
	cluster list.
*/


MRModel::MRModel() :
	pushMethod(0),
	numPushMethods(2),
	lastT(0.0),
	lastComplexity(1.0),
	showMeta(false),
	showFaces(true),
	colourFaces(false),
	currentFaces(0),
	currentClusters(0),
	clustersPrepped(false),
	verticesPrepped(false),
	isMapped(false)
{
	colour = scPrimitive::sDefaultColour;
	points = 0;
	colours = 0;
}

MRModel::~MRModel()
{
}

Bool MRModel::sAvgClusColours = false;

Void MRModel::AddContraction(
				Int				child0, 
				Int				child1,
				Int				face0,
				Int 			face1,
	 			GCLReal			error,
	 			const Point 	&p,
	 			Int				delta
	 		)
{
	MRVertex	newVtx;
	
	if (vertices.NumItems() == 0)
	// if this is first contraction, add leaf contractions for all model points
	// also set up vertex face info list
	{
		Int		i;
		
		vertices.SetSize(points->NumItems());
		vertices.ClearTo(newVtx);
		vtxFlags.SetSize(points->NumItems() * 2 - 1);
		vtxFlags.ClearTo(Flags8());

		// make the leaf vertices active
		for (i = 0; i < vertices.NumItems(); i++)
			vtxFlags[i].Set(MRV_Active);
		
		for (i = 0; i < faces.NumItems(); i++)
			vtxFlags[i].Set(MRV_FaceActive);
	}

	points->Append(p);

	if (delta)
		points->Last() += points->Item(child0);

	// MMF format uses an in-place numbering scheme (a contracted
	// vertex replaces its left child in a vertex array) so we
	// must walk up to the roots of the indicated vertices to
	// find the vertex that's really being referred to.

	while (!vertices[child0].IsRoot())
		child0 = vertices[child0].parent;
	while (!vertices[child1].IsRoot())
		child1 = vertices[child1].parent;

	// Set the contraction!
	newVtx.SetContraction(vertices, vertices.NumItems(), vtxFlags,
		child0, child1, face0, face1, error);
	vertices.Append(newVtx);

	Assert(vtxFlags[vertices.NumItems() - 1].IsSet(MRV_Inside), "flag not set");
	
	Assert(vertices.NumItems() == points->NumItems(), "point/vertex mismatch");
}

Void MRModel::AddCluster(
				Int         clusLeft,
				Int         clusRight,
				Point       &clusOrig,
				VecTrans    &clusAxis,
				Vector      &clusFitNormal,
				GCLReal     clusD,
				Point       &clusMin,
				Point       &clusMax,
				Int			clusID
			)
{
	Point           obbOrig;
	Vector          scales;
	Int				i;
	
	clusLeft--;
	clusRight--;

	if (clusters.NumItems() == 0)
	{
		MRV_COUT << "adding first cluster: preallocating space for " << 
			faces.NumItems() - 1 << " clusters" << endl;
		clusters.SetSize(faces.NumItems() - 1);
		cOffset = 2 * (faces.NumItems() - 1);
		clustersActive.SetSize(faces.NumItems() * 2 - 1);
		for (i = 0; i < faces.NumItems(); i++)
			clustersActive[i] = 1;
		
		MRV_COUT << "done." << endl;
	}

	Assert(clusLeft < clustersActive.NumItems() && clusRight < clustersActive.NumItems(),
		"Clusters not in tree order in cmf file.");

	FaceCluster &newCluster = Cluster(clusID);
	
	if (clusID & 8191 == 0)
		MRV_COUT << "adding cluster " << clusID << endl;
		
	newCluster.child[0] = clusLeft;
	newCluster.child[1] = clusRight;

	// convert to our storage format.

	// Garland's OBB axes are not unit length, and may not be right
	// handed. (The smallest evec may have been flipped to point in the
	// same direction as the sum-area normal).

	scales[0] = len(clusAxis[0]);
	scales[1] = len(clusAxis[1]);
	scales[2] = len(clusAxis[2]);
	// correct if left-handed
	if (det(clusAxis) < 0)
	{
		// flip y axis to correct, 'cause we want to keep z pointing same way
		// as normal
		scales[1] = -scales[1];
		Swap(clusMin[1], clusMax[1]);
	}
	clusAxis[0] /= scales[0];
	clusAxis[1] /= scales[1];
	clusAxis[2] /= scales[2];
	clusMin *= scales;
	clusMax *= scales;
	newCluster.SetAxes(clusAxis);

	// transform origin into bbox coords
	obbOrig = clusAxis * clusOrig;
	clusMin += obbOrig;
	clusMax += obbOrig;

	newCluster.min = clusMin;
	newCluster.max = clusMax;

	for (i = 0; i < 6; i++)
		newCluster.sideArea[i] = 255;

	Assert(clusMin[0] <= clusMax[0], "bad x bb");
	Assert(clusMin[1] <= clusMax[1], "bad y bb");
	Assert(clusMin[2] <= clusMax[2], "bad z bb");

	clustersActive[clusID] = 1;

	if (clusLeft >= 0)
	{
		Assert(clustersActive[clusLeft], "left child not clusterable");
		clustersActive[clusLeft] = 0;
	}
	if (clusRight >= 0)
	{
		Assert(clustersActive[clusRight], "right child not clusterable");
		clustersActive[clusRight] = 0;
	}
}

static GCLReal ProjectedLength(const Point &p1, const Point &p2,
							   const Transform &projMat)
{
	HPoint	hp1(p1, 1.0), hp2(p2, 1.0);
	Vector	c1, c2;
	Coord	result;

	c1[0] = dot(projMat[0], hp1);
	c1[1] = dot(projMat[1], hp1);
	c1[2] = dot(projMat[3], hp1);

	c2[0] = dot(projMat[0], hp2);
	c2[1] = dot(projMat[1], hp2);
	c2[2] = dot(projMat[3], hp2);

	result = proj(c1) - proj(c2);

	return(sqrlen(result));
}


Void MRModel::AdaptLength(
				GCLReal				threshold,
				const Transform		&M,			// model matrix
				const Transform		&P,			// projection matrix
				GCLReal				timeLimit	///< currently ignored.
			)
/*!
	Given model matrix M and projection matrix P, adapt so that
	the error in screen space is below threshold.
*/ 
{
    Transform	PM = xform(P, M);
	Int			i;

    // Square the threshold so we can avoid all those sqrt's
	threshold = sqr(threshold);

	// iterate over active vertices -- brute force for now...
	for (i = 0; i < vertices.NumItems(); i++)
		if (vtxFlags[i].IsSet(MRV_Active))
		{
			Bool		canExpand = false;
			Bool		canContract = false;
			MRVertex	&v = vertices[i];
			
			if (v.CanExpand(i, vtxFlags))
			{
				if (threshold < ProjectedLength(points->Item(i), 
							points->Item(v.child[0]), PM))
					canExpand = true;
				else if (threshold < ProjectedLength(points->Item(i), 
							points->Item(v.child[1]), PM))
					canExpand = true;
			}

			if (!v.IsRoot() && vertices[v.parent].CanContract(v.parent, vtxFlags))
			{
				MRVertex	&vp = vertices[v.parent];
				
				if (threshold > ProjectedLength(points->Item(v.parent), 
							points->Item(vp.child[0]), PM))
					canContract = true;
				else if (threshold > ProjectedLength(points->Item(v.parent), 
							points->Item(vp.child[1]), PM))
					canContract = true;
			}

			if (canExpand)
				v.Expand(i, vtxFlags);
			else if (canContract)
				vertices[v.parent].Contract(v.parent, vtxFlags);
			
			if (canExpand && canContract)
				MRV_COUT << "YOW!  Tricky situation." << endl;
		}

	AdjustLeafFaces();
}

Void MRModel::AdaptError(GCLReal t)
{
	Int		i;
	
	MRV_COUT << "Adapting..." << endl;
	
    if (t < lastT)
    {
		// Search for expansions


		MRV_COUT << "Expanding: " << endl;

		for (i = vertices.NumItems() - 1; i >= 0; i--)
		{
			MRVertex	&v = vertices[i];

			if (v.CanExpand(i, vtxFlags) && v.error >= t)
				v.Expand(i, vtxFlags);
		}
    }
    else
    {
		// Search for contractions
 
		MRV_COUT << "Contracting: " << endl;

		for (i = 0; i < vertices.NumItems(); i++)
		{
			MRVertex	&v = vertices[i];

			if (v.CanContract(i, vtxFlags) && v.error <= t)
				v.Contract(i, vtxFlags);
		}
    }

	AdjustLeafFaces();
    lastT = t;
	MRV_COUT << "model faces: " << currentFaces << endl;
}

Void MRModel::AdaptFaces(Int targetFaces)
{
	Int		i;
	
	MRV_COUT << "Adapting to " << targetFaces << endl;
	
    if (targetFaces > currentFaces)
    {
		// Search for expansions until we have enough faces

		MRV_COUT << "Expanding: " << flush;

		for (i = vertices.NumItems() - 1; targetFaces > currentFaces && i >= 0; i--)
		{
			MRVertex	&v = vertices[i];

			if (v.CanExpand(i, vtxFlags))
			{
				v.Expand(i, vtxFlags);
				if (v.face[0] >= 0)
					currentFaces++;
				if (v.face[1] >= 0)
					currentFaces++;
			}
		}
    }
    else if (targetFaces < currentFaces)
    {
		// Search for contractions
 
		MRV_COUT << "Contracting: " << flush;

		for (i = 0; (targetFaces < currentFaces) && i < vertices.NumItems(); i++)
		{
			MRVertex	&v = vertices[i];

			if (v.CanContract(i, vtxFlags))
			{
				v.Contract(i, vtxFlags);
				if (v.face[0] >= 0)
					currentFaces--;
				if (v.face[1] >= 0)
					currentFaces--;
			}
		}
    }

	AdjustLeafFaces();

	MRV_COUT << "Model faces: " << currentFaces << endl;
}

Void MRModel::AdaptClusters(Int targetClusters)
{
	Int		i;
	
	MRV_COUT << "Adapting to " << targetClusters << endl;
	
	if (clustersActive.NumItems() == 0)
	// don't have an active cluster list, so better create one now.
	{
		MRV_COUT << "Creating active list" << endl;
		currentClusters = rootClusters.NumItems();
		clustersActive.SetSize(faces.NumItems() * 2 - 1);
		for (i = 0; i < clustersActive.NumItems(); i++)
			clustersActive[i] = 0;
//		firstActiveClus = clustersActive.NumItems();
		for (i = 0; i < rootClusters.NumItems(); i++)
		{
//			firstActiveClus = Min(rootClusters[i], firstClusActive);
			clustersActive[rootClusters[i]] = 1;
		}
	}

    if (targetClusters > currentClusters)
    {
		// Search for expansions until we have enough faces

		MRV_COUT << "Expanding: " << flush;

		// move through from end of log, expanding active faces
		for (i = clustersActive.NumItems() - 1; (targetClusters > currentClusters) && i >= 0; i--)
			if (clustersActive[i] && !IsLeafCluster(i))
			{
//				firstClusActive = i;
				clustersActive[i] = 0;
				clustersActive[Cluster(i).child[0]] = 1;
				clustersActive[Cluster(i).child[1]] = 1;
				currentClusters++;
			}
    }
    else if (Max(rootClusters.NumItems(), targetClusters) < currentClusters)
    {
		// Search for contractions
 
		MRV_COUT << "Contracting: " << flush;

		// move through from start of log, contracting active faces
		for (i = faces.NumItems(); (targetClusters < currentClusters) 
			&& i < clustersActive.NumItems(); i++)
		{
			Int left = Cluster(i).child[0];
			Int right = Cluster(i).child[1];
			
			if (clustersActive[left])
			{
//				firstClusActive = i;
				clustersActive[i] = 1;
				clustersActive[left] = 0;
				clustersActive[right] = 0;
				currentClusters--;
			}
		}
    }

	MRV_COUT << "model clusters: " << currentClusters << endl;
}

Void MRModel::AdaptComplexity(GCLReal complexity)
/*!	We take # faces = complexity^2 * total faces.
	The justification for this is that if we define complexity
	as detail in a single dimension, for surfaces the number
	of faces will go up as the square of the complexity.
*/
{
	if (complexity != lastComplexity)
	{
		if (vertices.NumItems() > 0)
			AdaptFaces(sqr(complexity) * faces.NumItems());
		if (clusters.NumItems() > 0)
			AdaptClusters(sqr(complexity) * faces.NumItems());
		lastComplexity = complexity;
	}	
}

Void MRModel::AdjustLeafFaces()
/*! Adjusts the model's face indexes to match the current active
	simplified vertices. We do this as a separate operation to
	the vertex split/edge collapse operations to avoid having
	to store adjacency information.
*/
{	
	if (vertices.NumItems() == 0)
		currentFaces = faces.NumItems();
	else
	{
		Int		i, numActiveFaces = 0;
	
		for (i = 0; i < vtxFaces.NumItems(); i++)
			if (vtxFlags[i].IsSet(MRV_FaceActive))
			{
				vtxCodes[i].AdjustVertices(vertices, vtxFlags, vtxFaces[i].v, faces[i].v);
				numActiveFaces++;
			}

		currentFaces = numActiveFaces;
	}
}

Void MRModel::DumpFCH(StrConst filename)
/// Dump the cluster hierarchy as a .hier file.
{
	Int			i, group = 0, lastActive;
	ofstream	s;
	IndexList	stack;
	
	s.open(filename);

	for (i = 0; i < clustersActive.NumItems(); i++)
	{
		if (clustersActive[i] != group)
		{
			group = clustersActive[i];
			s << "g " <<  group << endl;
		}
		s << "e " << i;
		if (!IsLeafCluster(i))
			s << ' ' << Cluster(i).child[0] << ' ' << Cluster(i).child[1];
		s << endl;
	}

	lastActive = -1;
	stack.Clear();

	for (i = 0; i < rootClusters.NumItems(); i++)
		stack.Push(rootClusters[i]);

	while (stack.NumItems() != 0)
	{
		i = stack.Last();
		stack.Pop();
		if (clustersActive[i])
		{
			if (lastActive >= 0)
				s << "l " << lastActive << ' ' << i << endl;
			lastActive = i;
		}
		else if (!IsLeafCluster(i))
		{
			stack.Push(Cluster(i).child[1]);
			stack.Push(Cluster(i).child[0]);
		}
	}

	s.close();
}

#ifdef UNFINISHED
Void MRModel::CropClusters(Int numClusters)
{
	Int				i;
	FaceIdxArray	newFaces(faces);
	
	for (i = 0; i < rootClusters.NumItems(); i++)
		ReorderLeafClusters(rootClusters[i], newFaces);

	faces.SwapWith(newFaces);

	for (i = faces.NumItems(); i < clustersActive.NumItems(); i++)
	{
		if (clusters[i].child[0] < n)
			clusters.child[0] = 
	}
		
}
#endif

Int MRModel::ReorderLeafClusters(Int ic, FaceIdxArray &newFaces)
/** Reorder cluster faces consecutively.

	We reorder the faces so that the faces belonging to a particular
	cluster come consecutively.  This allows us to truncate the cluster
	hierarchy;

	It also provides for the possibility of placing start/end ID fields
	in the cluster to avoid descending the tree to figure out the faces
	belonging to a particular cluster.
*/
{
	if (IsLeafCluster(ic))
	{
		Int		newID = newFaces.NumItems();

		newFaces.Append(faces[ic]);

		return(newID);
	}
	else
	{
		FaceCluster &clus = Cluster(ic);

		clus.child[0] = ReorderLeafClusters(clus.child[0], newFaces);
		clus.child[1] = ReorderLeafClusters(clus.child[1], newFaces);

		return(ic);
	}
}

Void MRModel::PrepareClusters()
{
	Int		i, j;
	Vector	areaNormal;
		
	if (clustersPrepped)
		return;
		
	rootClusters.Clear();

	for (i = 0; i < clustersActive.NumItems(); i++)
		if (clustersActive[i])
			rootClusters.Append(i);
		
	// accumulate area normals up from leaves
	for (i = faces.NumItems(); i < clustersActive.NumItems(); i++)
	{
		FaceCluster &fc = Cluster(i);
		
		fc.areaNormal = vl_0;
		fc.totArea = 0;
		
		for (j = 0; j < 2; j++)
			if (IsLeafCluster(fc.child[j]))
			{
				areaNormal = FaceAreaNormal(fc.child[j]);
				fc.areaNormal += areaNormal;
				fc.totArea += len(areaNormal);
			}
			else
			{
				fc.areaNormal += Cluster(fc.child[j]).areaNormal;
				fc.totArea += Cluster(fc.child[j]).totArea;
			}
	}

	currentClusters = rootClusters.NumItems();

	MRV_COUT << "there are " << rootClusters.NumItems()
		 << " root clusters, total clusters = " 
		 << clustersActive.NumItems() << ", original faces = " << faces.NumItems() << endl;

	clustersPrepped = true;
}

Void MRModel::PrepareVertices()
{
	GCLVec		weights(vertices.NumItems(), vl_0);
	Int			i, j, k;
	Int			maxHeight, minHeight, height;
	TreeCodes	treeCodes(vertices.NumItems());
	VectorList	faceANs(faces.NumItems());

	if (verticesPrepped)
		return;

	maxHeight = 0;
	minHeight = 32767;
	maxErr = 0;
	rootVertices.Clear();
	
	// find all root vertices, and set up the hierarchies
	// corresponding to each one.
	for (i = 0; i < vertices.NumItems(); i++)
	{
		maxErr = Max(vertices[i].error, maxErr);

		if (vertices[i].IsRoot())
		{
			height = 0;
			MRSetupHierarchy(vertices, i, treeCodes, height, 0, 0);
			maxHeight = Max(height, maxHeight);
			minHeight = Min(height, minHeight);
			rootVertices.Append(i);
		}	
	}

	MRV_COUT << rootVertices.NumItems() << " root vertices" << endl;
	
	for (i = 0; i < vertices.NumItems(); i++)
	{
		vertices[i].areaNormal = vl_0;
//		vertices[i].totArea = 0;
	}

	// Set up face info (area, normal, etc.)
	vtxCodes.SetSize(faces.NumItems());	
	vtxFaces = faces;

	for (i = 0; i < vtxFaces.NumItems(); i++)
	{
		faceANs[i] = FaceAreaNormal(i);
				
		for (j = 0; j < 3; j++)
		{
			Assert(vertices[faces[i].v[j]].IsLeaf(), "Not at leaf vertices.");

			vtxCodes[i].treeCode[j] = treeCodes[faces[i].v[j]];
			vertices[faces[i].v[j]].areaNormal += faceANs[i];
			weights[faces[i].v[j]] += 1.0;
		}
	}

	// calculate vertex normals & equiv. areas. for leaf vertices
	for (i = 0; i < vertices.NumItems(); i++)
		if (weights[i] > 0.0)
		{
			GCLReal w = weights[i];
			
			vertices[i].areaNormal /= w;
//			vertices[i].totArea = len(vertices[i].areaNormal);
		}

	// calculate aggregate properties for non-leaf vertices.
	for (i = 0; i < rootVertices.NumItems(); i++)
		vertices[rootVertices[i]].PullProps(vertices);

	MRV_COUT << "there are " << rootVertices.NumItems()
		 << " root vertices, min/max vertex tree height = " 
		 << minHeight << '/' << maxHeight
		 << ", original vertices = " << points->NumItems() << endl;

	SimplestModel();

	verticesPrepped = true;
}

Void MRModel::PrepareModel()
{
	MRV_COUT << "processing MR model" << endl;

	if (clusters.NumItems() > 0) 
		PrepareClusters();
	if (vertices.NumItems() > 0)
		PrepareVertices();

	MRV_COUT << "done" << endl;
}

enum MRB_Flags
{
	MRB_HasVH 	= 0x0001,
	MRB_HasFH 	= 0x0002,
	MRB_Float	= 0x0004
};

struct MRB_Header
{
	Int			version;
	Flags32		flags;
};

#ifdef MR_MEM_MAP

// XXX shift to CL
Byte *MemMap(StrConst filename, Int &size)
{
	Int		fd;
	Byte	*result = 0;
	
	fd = open(filename, O_RDWR);
	
	if (fd != -1)
	{
		Int			fsize;
		struct stat	fstats;
		
		fstat(fd, &fstats);
		fsize = fstats.st_size;

		result = (Byte*) mmap(0,  fsize, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);

		size = fsize;
		if (result == MAP_FAILED)
			_Error("MemMap failed ");
		close(fd);
	}

	return(result);
}

Void ReadArray(NBaseArray &a, Byte* &p, Int &size)
{
	Int		elts, chunk;

	elts = *((Int*) p);
	p += sizeof(Int);

	a.Attach(p, elts, true);
	chunk = a.EltSize() * elts;
	p += chunk;
	size -= chunk;

	Assert(size >= 0, "End of mapped memory chunk");
}

template <class T> Void ReadArray(Array<T> &a, Byte* &p, Int &size)
{
	Int		elts, chunk;

	elts = *((Int*) p);
	p += sizeof(Int);

	a.Attach((T*) p, elts, true);
	chunk = sizeof(T) * elts;
	p += chunk;
	size -= chunk;

	Assert(size >= 0, "End of mapped memory chunk");
}

#ifdef CL_SGI_INST
#pragma instantiate Void ReadArray(Array<Int> &a, Byte* &p, Int &size)
#endif
#endif

Bool MRModel::ParseBinary()
{
	Bool		result = false;
	Int			i, j;
	FileName	binaryFile(modelFile);
	
	binaryFile.SetPath(modelFile.GetPath());
	binaryFile.SetExtension("mrb");
	MRV_COUT << "checking for " << binaryFile.GetPath() << endl;
		
	if (modelFile.IsReadable())
		do
		{
			MRV_COUT << "FOUND BINARY MRB" << endl;

#ifdef MR_MEM_MAP
			Byte		*p;
			MRB_Header	*header;
			Int			size = 0;

			MRV_COUT << "memmapping cluster file" << endl;
			p = MemMap(modelFile.GetPath(), size);

			MRV_COUT << "memmapped " << size << " bytes" << endl;
			header = (MRB_Header*) p;		
			p += sizeof(MRB_Header);
			size -= sizeof(MRB_Header);
			Assert(size >= 0, "End of mapped memory chunk");
			if (header->version > 1)
			{
				_Warning("Newer version");
			}

			MRV_COUT << "version " << header->version 
				<< ", flags " << header->flags << endl;

			ReadArray(*points, p, size);
			MRV_COUT << points->NumItems() << " points" << endl;

			ReadArray(faces, p, size);
			MRV_COUT << faces.NumItems() << " faces" << endl;

			if (header->flags.IsSet(MRB_HasVH))
			{
				ReadArray(rootVertices, p, size);
				MRV_COUT << rootVertices.NumItems() << " root vertices" << endl;
				ReadArray(vertices, p, size);
				MRV_COUT << vertices.NumItems() << " vertices" << endl;
				ReadArray(vtxCodes, p, size);
				MRV_COUT << vtxCodes.NumItems() << " codes" << endl;
			}

			if (header->flags.IsSet(MRB_HasFH))
			{
				ReadArray(rootClusters, p, size);
				MRV_COUT << rootClusters.NumItems() << " root clusters" << endl;
				ReadArray(clusters, p, size);
				MRV_COUT << clusters.NumItems() << " clusters" << endl;
			}
			
			result = true;
			isMapped = true;
#else
			FILE		*file;
			MRB_Header	header;
			
			file = binaryFile.FOpen("r+b");
			if (!file)
			{
				_Warning("Couldn't open file");
				break;
			}

			do {
			if (sizeof(MRB_Header) != fread(&header, 1, sizeof(MRB_Header), file))
			{
				_Warning("couldn't read header");
				break;
			}
			
			if (header.version > 1)
			{
				_Warning("Newer version");
			}

			MRV_COUT << "version " << header.version 
				<< ", flags " << header.flags << endl;
						
			if (points->FRead(file)) break;
			MRV_COUT << points->NumItems() << " points" << endl;

			if (faces.FRead(file)) break;
			MRV_COUT << faces.NumItems() << " faces" << endl;

			if (header.flags.IsSet(MRB_HasVH))
			{
				if (rootVertices.FRead(file)) break;
				MRV_COUT << rootVertices.NumItems() << " root vertices" << endl;
				if (vertices.FRead(file)) break;
				MRV_COUT << vertices.NumItems() << " vertices" << endl;
				if (vtxCodes.FRead(file)) break;
				MRV_COUT << vtxCodes.NumItems() << " codes" << endl;
			}
			if (header.flags.IsSet(MRB_HasFH))
			{
				if (rootClusters.FRead(file)) break;
				MRV_COUT << rootClusters.NumItems() << " root clusters" << endl;
				if (clusters.FRead(file)) break;
				MRV_COUT << clusters.NumItems() << " clusters" << endl;
			}
			
			result = true;
			} while (false);

			if (ferror(file))
			{
				perror("error reading MRB file");
				break;
			}
#endif

			if (vertices.NumItems() > 0)
			{
				Flags8	fa;
				fa.Set(MRV_FaceActive | MRV_Inside);
		
				MRV_COUT << "starting contraction prep" << endl;
				vtxFaces = faces;
				vtxFlags.SetSize(Max(vertices.NumItems(), faces.NumItems()));
				vtxFlags.ClearTo(fa);
				for (i = 0; i < faces.NumItems(); i++)
					for (j = 0; j < 3; j++)
					{
						vtxFlags[faces[i].v[j]].Set(MRV_Active);
						vtxFlags[faces[i].v[j]].UnSet(MRV_Inside);
					}
				AdjustLeafFaces();	// XXX store ms indexes to avoid?
				verticesPrepped = true;
				MRV_COUT << "done" << endl;
			}
			
			if (clusters.NumItems() > 0)
			{
				cOffset = 2 * (faces.NumItems() - 1);

				clustersPrepped = true;
			}
		}
		while (false);
	
	return(result);
}

Void MRModel::WriteBinary()
{
	FileName	binaryFile;

	if (isMapped || (clusters.NumItems() == 0 && vertices.NumItems() == 0))
		return;

	binaryFile.SetDir(".");
	binaryFile.SetFile(modelFile.GetFile());
	binaryFile.SetExtension("mrb");

	cerr << "Writing binary to " << binaryFile.GetPath() << endl;
	
	do
	{
		FILE		*file;
		MRB_Header	header;
		
		// XXX needs error proofed

		header.version = 1;
		if (vertices.NumItems() > 0)
			header.flags.Set(MRB_HasVH);
		if (clusters.NumItems() > 0)
			header.flags.Set(MRB_HasFH);
#ifdef GCL_FLOAT
		head.flags.Set(MRB_Float);
#endif
	
		if (!points)
			_Error("Model has no points?");

		MRV_COUT << "WRITING BINARY MRB FILE" << endl;
		file = binaryFile.FOpen("wb");
		if (!file)
			break;
		
		do 
		{
			MRV_COUT << "ACTIVE: " << currentClusters << endl;

			if (sizeof(MRB_Header) != fwrite(&header, 1, sizeof(MRB_Header), file)) break;
			MRV_COUT << "points: " << points->NumItems() << endl;
			if (points->FWrite(file)) break;
			MRV_COUT << "faces: " << faces.NumItems() << endl;
			if (faces.FWrite(file)) break;

			if (vertices.NumItems() > 0)
			{
				MRV_COUT << "root vertices: " << rootVertices.NumItems() << endl;
				if (rootVertices.FWrite(file)) break;
				MRV_COUT << "vertices: " << vertices.NumItems() << endl;
				if (vertices.FWrite(file)) break;
				MRV_COUT << "codes: " << vtxCodes.NumItems() << endl;
				if (vtxCodes.FWrite(file)) break;
			}

			if (clusters.NumItems() > 0)
			{
				MRV_COUT << "root clusters: " << rootClusters.NumItems() << endl;
				if (rootClusters.FWrite(file)) break;
				MRV_COUT << "clusters: " << clusters.NumItems() << endl;
				if (clusters.FWrite(file)) break;
			}
		}
		while (false);

		if (ferror(file))
			perror("MRB write");
		fclose(file);
		MRV_COUT << "done." << endl;
	} while (false);
}

Bool MRModel::ParseText()
{
	String		func;
	Point		vtx;
	Int			numVertices;
	Int			faceIdx = 0, vtxOffset, delta;

	Int			child[2], depFace[2];
	Int         vtxId;       
	GCLReal     cost;       
	Point		place;

	Int			clusID, clusLeft, clusRight;
	Point		clusOrig;
	VecTrans	clusAxis;
	Vector		clusFitNormal;
	Point		clusMin, clusMax;
	GCLReal		clusD;

	ifstream 	s;

	vtxOffset = -1; // default is 1-based indexes
	delta = 0; 		// original mmf format had no delta-encoding
	clusID = -1;
	child[0] = -1;
	clusLeft = -1;
	cost = 0.01;
	
	if (!points)
		points = new PointList;
	
	s.open(modelFile.GetPath());
	if (!s)
	{
		cerr << "Cannot access " << modelFile.GetPath() << endl;
		return(false);
	}

	numVertices = 0;
	
	while (s)
	{
		if (func.ReadWord(s))
		{
			if (func == "#$cost" || func == "ve")
				s >> vtxId >> cost;
			else if (func[0] == '#')
				;
			else if (func == "v")
			{
				s >> vtx[0] >> vtx[1] >> vtx[2];
				points->Append(vtx);
				numVertices++;
			}
			else if (func == "f" || func == "t")
			{
				Int		a, b, c;

				s >> a >> b >> c;

				faces.Add(1);
				faces.Last().v[0] = a - 1;
				faces.Last().v[1] = b - 1;
				faces.Last().v[2] = c - 1;
			}
			else if (func == "set")
			{				
				func.ReadWord(s);
				if (func == "vertex_correction")
				{
					// add vertex_correction to indexes to get 1-based
					// indexing. We use 0-based indexing, hence the
					// -1.
					s >> vtxOffset;
					vtxOffset = vtxOffset - 1;
				}
				else if (func == "delta_encoding")
					s >> delta;
				else
					cerr << "(ParseMMF) *** Ignoring unknown set variable: "
						 << func << endl;
			}
			else if (func == "%SMF" || func == "#$SMF")
			{
				GCLReal version;
				s >> version;
				if (version < 1.0)
					vtxOffset = 0; // version 0 had 0-based offsets
			}

// === vertex-cluster stuff
			else if (func == "v%")
			// a vertex contraction
			{
				Char	gobbleBracket;
				
				// add previous contraction
				if (child[0] >= 0)
					AddContraction(child[0] + vtxOffset, child[1] + vtxOffset, 
					    depFace[0] - 1, depFace[1] - 1, cost, place, delta);

				// read in start of new contraction
				ChompWhiteSpace(s);
				if (s.peek() == '(')
					s >> gobbleBracket;
				s >> child[0] >> child[1];
				ChompWhiteSpace(s);
				if (s.peek() == ')')
					s >> gobbleBracket;
				s >> place[0] >> place[1] >> place[2];
				faceIdx = 0;
				depFace[0] = depFace[1] = 0;
				cost += 0.01;	// assume constant cost per contraction if
								// none is specified in the file.
			}
			else if (func == "f-")
			{
				// dependent face id of current contraction
				s >> depFace[faceIdx];
				faceIdx ^= 1;
			}

// === face-cluster stuff
			else if (func == "f^")
			{
				if (clusID < 0)
					clusID = faces.NumItems(); // first cluster!

				// add previous cluster record
				if (clusLeft > 0)
				{
					AddCluster(clusLeft, clusRight, clusOrig, clusAxis,
						clusFitNormal, clusD, clusMin, clusMax, clusID++);
				}
				s >> clusLeft >> clusRight;
			}
			else if (func == "fo")
				s >> clusOrig[0] >> clusOrig[1] >> clusOrig[2];
			else if (func == "fe")
			{
				s >> clusAxis[0][0] >> clusAxis[0][1] >> clusAxis[0][2];
				s >> clusAxis[1][0] >> clusAxis[1][1] >> clusAxis[1][2];
				s >> clusAxis[2][0] >> clusAxis[2][1] >> clusAxis[2][2];
			}
			else if (func == "fd")
				s >> clusD;
			else if (func == "fn")
				s >> clusFitNormal[0] >> clusFitNormal[1] >> clusFitNormal[2];
			else if (func == "fx")
			{
				s >> clusMin[0] >> clusMin[1] >> clusMin[2];
				s >> clusMax[0] >> clusMax[1] >> clusMax[2];
			}
			else if (func == "fc")
				;	// cost -- ignore it.

// unrecognized!
			else
				cerr << "(ParseMMF) *** Ignoring unknown token: " << func << endl;

			func.ReadLine(s); // ignore rest of line
		}
	}

	s.close();

	if (child[0] >= 0)
		AddContraction(child[0] + vtxOffset, child[1] + vtxOffset, 
					   depFace[0] - 1, depFace[1] - 1, cost, place, delta);

	if (clusLeft > 0)
		AddCluster(clusLeft, clusRight, clusOrig, clusAxis,
			clusFitNormal, clusD, clusMin, clusMax, clusID++);

	if (clusID > 0)
		clustersActive.SetSize(clusID);

	MRV_COUT << "Read " << numVertices << " vertices and " << faces.NumItems()
		 << " faces." << endl;

	return(true);
}

Bool MRModel::Parse(StrConst filename)
{
	Bool	result;
	
	if (!points)
		points = new PointList;

	modelFile.SetPath(filename);

	if (modelFile.GetExtension() == "mrb")
		result = ParseBinary();
	else 
		result = ParseText();

	if (!result)
		return(false);
			
	MRV_COUT << "There are " << vertices.NumItems() << " vertex contractions, "
		<< clusters.NumItems() << " face clusters." << endl;
		
	MRV_COUT << "MEMORY:" << endl;
	if (points)
		MRV_COUT << "  points     " << sizeof(Point) * points->NumItems() / 1024.0 << endl;
	if (colours)
		MRV_COUT << "  colours    " << sizeof(Colour) * colours->NumItems() / 1024.0 << endl;
	MRV_COUT << "  indexes    " << sizeof(FaceIdx) * faces.NumItems() / 1024.0 << endl;
	MRV_COUT << "  vtxFaces   " << sizeof(FaceIdx) * vtxFaces.NumItems() / 1024.0 << endl;
	MRV_COUT << "  vtxCodes   " << sizeof(MRCodes) * vtxCodes.NumItems() / 1024.0 << endl;
	MRV_COUT << "  vertices   " << sizeof(MRVertex) * vertices.NumItems() / 1024.0 << endl;
	MRV_COUT << "  clusters   " << sizeof(FaceCluster) * clusters.NumItems() / 1024.0 << endl;

	PrepareModel();
	
	return(true);
}

Void MRModel::DrawClusterFaces(Int ic, Renderer &r)
{
	if (IsLeafCluster(ic))
	{
		PointList 	&pts = *points;
		Int			*triIdx = faces[ic].v;
		Vector		normal;

		normal = FaceAreaNormal(ic);
		
		r	.Begin(scPrimitive::sRenderStyle)
			.N(norm(normal))
			.P(pts[triIdx[0]])
			.P(pts[triIdx[1]])
			.P(pts[triIdx[2]])
			.End();
	}
	else
	{
		FaceCluster &mrf = Cluster(ic);

		DrawClusterFaces(mrf.child[0], r);
		DrawClusterFaces(mrf.child[1], r);
	}	
}

Void MRModel::Draw(Renderer &r)
{
	Int		i;
	Int		*idx;
	Bool	hasVF = (vtxFaces.NumItems() > 0);
	
	if (colourFaces && clusterColours.NumItems() == 0)
		CreateClusterColours();

	if (showFaces)
	{
		if (colourFaces)
		{
			for (i = 0; i < clustersActive.NumItems(); i++)
				if (clustersActive[i])
				{
					r.C(clusterColours[i]);
					DrawClusterFaces(i, r);
				}
		}
		else
		{
			r.C(colour);
			for (i = 0; i < faces.NumItems(); i++)
			{
				if (hasVF)
				{
					if (!vtxFlags[i].IsSet(MRV_FaceActive))
						continue;
					idx = vtxFaces[i].v;
				}
				else
					idx = faces[i].v;
					
				// XXX shouldn't need norm()'s here, but MESA has prec. bug with
				// very small normals when GL_NORMALIZE... need workaround

				r.Begin(scPrimitive::sRenderStyle);
				if (vertices.NumItems() > 0)
				{
					if (colours)
						r
							.N(norm(vertices[idx[0]].areaNormal))
							.C((*colours)[idx[0]])
							.P((*points)[idx[0]])
							.N(norm(vertices[idx[1]].areaNormal))
							.C((*colours)[idx[1]])
							.P((*points)[idx[1]])
							.N(norm(vertices[idx[2]].areaNormal))
							.C((*colours)[idx[2]])
							.P((*points)[idx[2]]);
					else 
						r
							.N(norm(vertices[idx[0]].areaNormal))
							.P((*points)[idx[0]])
							.N(norm(vertices[idx[1]].areaNormal))
							.P((*points)[idx[1]])
							.N(norm(vertices[idx[2]].areaNormal))
							.P((*points)[idx[2]]);
				}
				else
				{
					Vector areaNormal;

					CalcTriAreaNormal(
						(*points)[idx[0]],
						(*points)[idx[1]],
						(*points)[idx[2]],
						areaNormal
					);

					r.N(norm(areaNormal));
					
					if (colours)
						r	.C((*colours)[idx[0]])
							.P((*points)[idx[0]])
							.C((*colours)[idx[1]])
							.P((*points)[idx[1]])
							.C((*colours)[idx[2]])
							.P((*points)[idx[2]]);
					else
					{
						r	.P((*points)[idx[0]])
							.P((*points)[idx[1]])
							.P((*points)[idx[2]]);
					}
				}
				r.End();
			}
		}
	}

	if (showMeta)
	{
		if (clustersActive.NumItems() > 0)
			DrawClusters(r);
		else if (vertices.NumItems() > 0)
			DrawVertices(r);
	}
}

Void MRModel::CreateClusterColours()
{
	if (clustersActive.NumItems() > 0)
	{
		Int		i, fc1, fc2;
		Colour	c;
		
		clusterColours.SetSize(clustersActive.NumItems());

		if (!sAvgClusColours)
		{
			for (i = 0; i < faces.NumItems(); i++)
				clusterColours[i] = HSVCol(drand48() * 360.0, 1.0, 1.0);
			for (i = faces.NumItems(); i < clusterColours.NumItems(); i++)
			{
				fc1 = Cluster(i).child[0];
				fc2 = Cluster(i).child[1];
				// XXX fix when we store face areaNormals
				if (IsLeafCluster(fc1) || IsLeafCluster(fc2) || (Cluster(fc1).totArea > Cluster(fc2).totArea))
					c = clusterColours[fc1];
				else
					c = clusterColours[fc2];
				
				clusterColours[i] = c;
			}
		}
		else
		{
			for (i = 0; i < faces.NumItems(); i++)
				clusterColours[i] = HSVCol(drand48() * 360.0, 1.0, 1.0);
			for (i = faces.NumItems(); i < clusterColours.NumItems(); i++)
			{
				c = clusterColours[Cluster(i).child[0]];
				c += clusterColours[Cluster(i).child[1]];
				c /= 2.0;
				clusterColours[i] = c;
			}
		}
	}
}

Void MRModel::DrawClusters(Renderer &r)
/// draw representations of face clusters
{
	Int		i;

	for (i = 0; i < clustersActive.NumItems(); i++)
		if (clustersActive[i])
			if (!IsLeafCluster(i))
			{
				if (!showFaces && colourFaces)
					r.C(Colour4(clusterColours[i], 0.5));
				else
					r.C(Colour4(colour, 0.5));
				Cluster(i).Draw(r);
			}
			else if (!showFaces)
			{
				if (colourFaces)
					r.C(clusterColours[i]);
				else
					r.C(cYellow);
				DrawClusterFaces(i, r);
			}
}

Void MRModel::DrawVertices(Renderer &r)
/// draw representations of simplification vertices
{
	Int		i;

	r.C(Colour4(cGreen, 0.5));
	
	for (i = 0; i < vertices.NumItems(); i++)
		if (vtxFlags[i].IsSet(MRV_Active))
		{
			r.N(vertices[i].areaNormal);
			DrawCircle(r, (*points)[i],
				norm(vertices[i].areaNormal), 
				sqrt(len(vertices[i].areaNormal) / vl_pi)
			);
		}
}

Void MRModel::SimplestModel()
{
	Int		i;

	AdaptClusters(0);
	for (i = 0; i < vertices.NumItems(); i++)
	{
		MRVertex	&v = vertices[i];
		
		if (v.CanContract(i, vtxFlags))
			v.Contract(i, vtxFlags);
	}

	lastT = maxErr;	
	lastComplexity = 0.0;
	AdjustLeafFaces();
}

Void MRModel::MostComplexModel()
{
	Int		i;

	AdaptClusters(faces.NumItems());
	
	for (i = vertices.NumItems() - 1; i >= 0; i--)
	{
		MRVertex	&v = vertices[i];
		
		if (v.CanExpand(i, vtxFlags))
			v.Expand(i, vtxFlags);
	}

	lastT = 0.0;
	lastComplexity = 1.0;
	AdjustLeafFaces();
}

Vector MRModel::FaceAreaNormal(Int faceIdx)
{
	Vector	an;
	Int		*triIdx = faces[faceIdx].v;

	CalcTriAreaNormal((*points)[triIdx[0]],
				(*points)[triIdx[1]],
				(*points)[triIdx[2]],
				an);

	return(0.5 * an);
}

// --- scMRModel methods ------------------------------------------------------


GCLReal scMRModel::complexity = 0.1;

scMRModel::scMRModel() : scPrimitive(pMRModel), model(), currentComplexity(0.0)
{
}

Void scMRModel::Draw(Renderer &r, SLContext *context)
{
	Colour 		*cPtr;
	ColourList	*cClrs;

	if (cPtr = SC_GET(Colour))
		model.colour = *cPtr;
	if (!model.colours && (cClrs = SC_GET(Colours)))
	{
		if (cClrs->NumItems() == model.points->NumItems())
			model.colours = cClrs;
		else
			cerr << "ignoring bad colours list in MRModel" << endl;
	}
	
	model.AdaptComplexity(complexity);
	model.Draw(r);
}

Void scMRModel::DecimateSelf(Decimator &dec)
{	
	if (dec.flags & DecIgnoreMRM)
		return;

	Int			i;
	FaceIdx		fi;
	Bool		changed, hasVF = (model.vtxFaces.NumItems() > 0);

	model.AdaptComplexity(complexity);

	dec.flags |= DecIsMRM;
	changed = true;

	for (i = 0; i < model.faces.NumItems(); i++)
	{
		if (hasVF)
		{
			if (!model.vtxFlags[i].IsSet(MRV_FaceActive))
				continue;
			else
				fi = model.vtxFaces[i];
		}
		else
			fi = model.faces[i];
			
		fi.v[0] += dec.pointsAccNum;
		fi.v[1] += dec.pointsAccNum;
		fi.v[2] += dec.pointsAccNum;

		dec.HandlePoly(3, fi.v, changed);
		changed = false;
		dec.decNum++;
	}

	dec.flags &= ~DecIsMRM;
}

Void scMRModel::SetComplexity(GCLReal comp)
{
	complexity = comp;
};

GCLReal scMRModel::GetComplexity()
{
	return(complexity);
};

#include "gcl/SceneLang.h"

scScenePtr ParseMMFFile(const Char *filename)
{
	scScenePtr 	result;
	scMRModel	*model;
	scPoints	*points;

	result = slBeginObject(filename);
		slPointList();
		points = (scPoints*) slCurrent();
		slObject("mr-model");
		model = (scMRModel*) slCurrent();
	slEndObject();

	model->model.points = points;
	model->model.Parse(filename);
	model->model.maxErr += 1e-8;
		
	return(result);
}

// --- ModelsFinder methods ---------------------------------------------------


Void MRModelsFinder::Start()
{
	scSceneAction::Start();
	models.Clear();
}

Void MRModelsFinder::Primitive(scPrimitive *sp)
{
	if (sp->PrimID() == pMRModel)
		models.Append(&((scMRModel*) sp)->model);
}
