/*
	File:		MRModel.h

	Function:	Implements edge-contraction multires polygonal model

	Author:		Andrew Willmott

	Copyright:	(c) 1997-1999
*/

#ifndef __MRModel__
#define __MRModel__

#include "gcl/MRVertex.h"
#include "gcl/MRFaceCluster.h"
#include "gcl/Renderer.h"

typedef Array<Byte> ActList;

class MRModel
/** Implements a multiresolution model. 

	Provides support for both edge-contraction and face-clustering
	kinds of polygonal model.
*/
{
public:
    MRModel();
	~MRModel();
	
	// Adaption routines
	Void			SimplestModel();
					///< Set to the simplest model possible
	Void			MostComplexModel();
					///< Set to the most complex model possible

	Void    		AdaptFaces(Int targetFaces);
					///< Adapt the model so it has 'targetFaces' faces
	Void    		AdaptClusters(Int targetClusters);
					///< Adapt the model so it has 'targetClusters' face clusters
	Void			AdaptComplexity(GCLReal complexity);
					///< complexity is [0, 1]; adapts model to totalFaces * sqr(c)
					/// clusters.

	Void    		AdaptLength(
							GCLReal				threshold,       
							const Transform 	&M,
							const Transform 	&P,      
							GCLReal				timeLimit = vl_inf
					);
					///< Adapt so screen error is constant.
					
	Void    		AdaptError(GCLReal threshold);
					///< adapt model so all contraction errors are smaller than threshold

	// Parsing
	Bool			Parse(StrConst filename);
	Bool			ParseBinary();
	Bool			ParseText();
	Void			WriteBinary();

	// Drawing
	Void			Draw(Renderer &r);
	
	// Internal utility routines
	Void			PrepareModel();
	Void			AdjustLeafFaces();
	
	Void			PrepareClusters();
	Void			PrepareVertices();

	Void			DrawClusterFaces(Int i, Renderer &r);
	Void			DrawClusters(Renderer &r);
	Void			DrawVertices(Renderer &r);

	Vector			FaceAreaNormal(Int faceIdx);
	Int				ReorderLeafClusters(Int ic, FaceIdxArray &newFaces);
	Void			CropClusters(Int numClusters);
	Void			CreateClusterColours();
	Void			DumpFCH(StrConst filename);
	
	Void			AddContraction(
							Int         child0,
							Int         child1,
							Int         face1,
							Int         face2,
							GCLReal     error,
							const Point &p,
							Int         delta
						);
						
	Void			AddCluster(
							Int         child0,
							Int         child1,
							Point       &centre,
							VecTrans    &axis,
							Vector      &fitNormal,
							GCLReal     d,
							Point       &clusMin,
							Point       &clusMax,
							Int			clusID
						);

// basic model
	PointList		*points;		///< vertex locations
	ColourList		*colours;		///< per-vertex colours, if they exist
	Colour			colour;			///< model colour
	FaceIdxArray	faces;			///< list of face vertex indicies

// vertex-contraction stuff
	Int				currentFaces;	///< number of faces in current model.
	IndexList		rootVertices;	///< root nodes of the vertex forest
	MRVertexes		vertices;		///< all vertex nodes
	MRCodesList		vtxCodes;		///< tree codes per face
	MRFlagsList		vtxFlags;		///< active vertices/faces
	FaceIdxArray	vtxFaces;		///< list of extra face info for MMF stuff
	Bool			verticesPrepped;

// face cluster stuff
	Int				currentClusters;///< number of clusters in current model.
	Bool			clustersPrepped;///< clusters have been preprocessed
	IndexList		rootClusters;	///< indices of root clusters
	FaceClusList	clusters;		///< all the face clusters
	ActList			clustersActive;	///< list of active clusters
	ColourList		clusterColours; ///< colours to show varying clusters
	NormalList		faceNormals;	///< area-weighted normals
	Int				cOffset;		///< cluster array offset (2f-2)
	
	Bool			IsLeafCluster(Int i)
					{ return(i < faces.NumItems()); };
	FaceCluster		&Cluster(Int i)
					{ return(clusters[cOffset - i]); };

	Float			maxErr;
	Int				pushMethod;
	Int				numPushMethods;
	
	Bool			showMeta;		///< show clusters/vertices
	Bool			showFaces;		///< show underlying model
	Bool			colourFaces;	///< colour faces according to face clusters

	GCLReal			lastT;
	GCLReal			lastComplexity;

	FileName		modelFile;		///< location of source file.

	Bool			isMapped;		///< model is memmapped
	static Bool		sAvgClusColours;///< parent's colour is average of children's
};


// --- Integration into GCL scenes ---------------------------------------------

#include "gcl/SceneObjects.h"

class scMRModel : public scPrimitive
/// A MR model that can be placed in a GCL scene.
{
public:

	scMRModel();

	Void			Draw(Renderer &r, SLContext *context);
	const Char		*Label() const { return "MR Model"; };

	Object			*Clone() const { return new scMRModel(*this); };

	Void			DecimateSelf(Decimator &dec);

	MRModel			model;

	static Void 	SetComplexity(GCLReal comp);
	static GCLReal	GetComplexity();
	
protected:
	static GCLReal	complexity;
	GCLReal			currentComplexity;
};

class MRModelsFinder : public scSceneAction
/// Builds a list of MR models in the scene.
{
public:
	Void				Start();
	Void				Primitive(scPrimitive *sp);

	PtrArray<MRModel>	models;

	Int					NumModels() const
						{ return(models.NumItems()); }
	MRModel				*Model(Int i) const
						{ return(models[i]); }
};


#endif
