/*
	File:		ParseMSDL.cc

	Function:	Parse Manchester Scene Description Language, using libmsdl.a

	Author:		Andrew Willmott

	Notes:		Adapted from the application.c file that comes with the msdl
				parser lib.
				
				Most of the MSDL model files don't seem to have consistently
				oriented polygons, although their spec does mention that
				polygons should follow the right-hand orientation rule. Oh well.
*/

#include "gcl/GCLConfig.h"
#ifdef GCL_MSDL

#include <stdlib.h>
// If I ruled the world, people who #define syscalls to be something else
// would be hunted down like dogs and shot. be warned: msdl_macros.h
// #defines free, which is why stdlib.h must be #included first.

extern "C" {
#include "msdl_types.h"
#include "msdl_macros.h"
#include "msdl_protos.h"
}
#include "gcl/SceneLang.h"
#include "gcl/VecUtil.h"

#ifdef DEBUG
#define VERBOSE
#endif

extern Void yyerror(char *errmess);

scScenePtr ParseMSDLFile(const Char *filename);

static Vec3d gUp;

Colour toColour(const Tspectra &c)
{
	return(Colour(c.r, c.g, c.b));
}

Point toPoint(const Tpt3 &p)
{
	return(Point(p.x, p.y, p.z));
}

scScenePtr ParseMSDLFile(const Char *filename)
{
	scScenePtr result = slBeginObject((char*) filename);
	FILE *instream;

	// msdl generally has ZUP instead of YUP, so correct...

	slTransform(Rotation(vl_y, vl_pi) * Rotation(vl_x, -vl_halfPi));

	instream = fopen(filename, "r");
	msdl(instream);

	slEndObject();

	return(result);
}


Void HandleProps(Tobject *obj)
{
	Colour c;

	if (obj->sprops.flags & EMmask)
		slEmittance(toColour(obj->sprops.emission));
	if (obj->sprops.flags & DRCmask)
		slColour(toColour(obj->sprops.drc));
}


/*****************************************************************************
 * APPLICATION.c this file contains some functions written by me and some
   functions written by you. It recurses through the object list and when it
   comes to a composite object it calls create_comp(). Upon finishing the
   composite it calls end_comp(). If it comes to a simple low level
   object such as spere, box, polygon it will call a function
   ceate_spere( Tobject* ), create_polygon( Tobject* )... etc....
   You must write these functions to do with teh objects what you want.
   If you are not bothered about composite objects just leave the create_comp()
   and end_comp() alone...

   After the object routines are skeletons of lights and viewing functions.
 ***************************************************************************/
#include <stdio.h>
#define _INCLUDE_XOPEN_SOURCE
#include <math.h>
#include "msdl_types.h"
#include "msdl_protos.h"

/* Function prototype */
Void print_obj_info( Tobject * );


/* ************************************************************* */
/* Error handler */

Void create_problem(char *probtext, int line_number)
{
  fprintf(stderr,"MSDL ERROR : %s at line %d\n",
	  probtext, line_number);
  exit(-1);
}


/*****************************************************************************
      ****    ******    ********  ********    ****    ********    *****
    **    **  **    **     **     **	    **    **     **     **     *
    **    **  ******       **     *****	    **	         **       ****
    **    **  **    **  *  **     **	    **    **     **     *     **
      ****    ******     ***      ********    ****	     **      *****
 *****************************************************************************/

/*===========================================================================*/
Void create_comp( Tobject* obj, Tmatrix acc_trans )
{
#ifdef VERBOSE
  printf("Creating a composite\n");
  printf("It's name is %s\n",obj->name );
#endif

  slBeginObject(obj->name);

#ifdef CLEANUP
  free( obj );
#endif
}
/*---------------------------------------------------------------------------*/


/*===========================================================================*/
Void end_comp( Void )
{
#ifdef VERBOSE
  printf("Ending composite\n");
#endif

  slEndObject();
}
/*---------------------------------------------------------------------------*/

/*===========================================================================*/


Void create_triangle( Tobject* obj, Tmatrix acc_trans )
{
#ifdef VERBOSE
  printf("Creating a triangle\n");
  print_obj_info( obj );
	printf(" It has %4d vertices \n",obj->spec.tri.vvnfn.no_verts );
	printf(" It has %4d vertex normals \n",obj->spec.tri.vvnfn.no_vnorms );
	printf(" It has %4d facet  normals \n",obj->spec.tri.vvnfn.no_fnorms );
#endif

	Int		i;
	Point	q;
	
	Expect(obj->spec.tri.vvnfn.no_verts == 3, "bogus triangle");
	HandleProps(obj);
	slPointList();

	for (i = 0; i < obj->spec.tri.vvnfn.no_verts; i++)
		slPoint(toPoint(transPt3(obj->spec.tri.vvnfn.data + i, acc_trans)));

	slPoly();

#ifdef CLEANUP
  free( obj->spec.tri.vvnfn.data );
  free( obj );
#endif
}
/*---------------------------------------------------------------------------*/

/*===========================================================================*/

Void create_polygon( Tobject* obj, Tmatrix acc_trans )
{
#ifdef VERBOSE
  printf("Creating a polygon\n");
  print_obj_info( obj );
  printf(" It has %4d vertices \n",obj->spec.polg.vvnfn.no_verts );
  printf(" It has %4d vertex normals \n",obj->spec.polg.vvnfn.no_vnorms );
  printf(" It has %4d facet  normals \n",obj->spec.polg.vvnfn.no_fnorms );
#endif

	Int		i;
	Mat4d	&t = (Mat4d&) acc_trans[0][0];
	Vec3d	p;
	
	HandleProps(obj);

	if (obj->spec.polg.vvnfn.no_vnorms > 0)
	{
		_Warning("poylgon vertex normals ignored");
		cerr << obj->spec.polg.vvnfn.no_vnorms << endl;
	}
	if (obj->spec.polg.vvnfn.no_fnorms > 0)
	{
		_Warning("poylgon face normals ignored");
		cerr << obj->spec.polg.vvnfn.no_fnorms << endl;
	}

	slPointList();

	for (i = 0; i < obj->spec.polg.vvnfn.no_verts; i++)
		slPoint(toPoint(transPt3(obj->spec.polg.vvnfn.data + i, acc_trans)));

	slPoly();

#ifdef CLEANUP
  free( obj->spec.polg.vvnfn.data );
  free( obj );
#endif
}
/*---------------------------------------------------------------------------*/

/*===========================================================================*/
Void create_sphere( Tobject* obj, Tmatrix acc_trans )
{
#ifdef VERBOSE
  printf("Creating a sphere\n");	
  print_obj_info( obj );
	printf(" Centre is "PCF" "PCF" "PCF" \n",
						obj->spec.sph.centre.x,
						obj->spec.sph.centre.y,
						obj->spec.sph.centre.z );
	printf(" It's radius is "PCF"\n", obj->spec.sph.radius );
#endif

	Vec3d	&centre = (Vec3d&) obj->spec.sph.centre;
	Vector	radius;
	Mat4d	&t = (Mat4d&) acc_trans[0][0];

	radius.MakeBlock(obj->spec.sph.radius * 2.0);
	slObject("sphere");
	slApply(xform(t, xform(Shift(centre), Scale(radius))));

#ifdef CLEANUP
  free( obj );
#endif
}
/*---------------------------------------------------------------------------*/

/*===========================================================================*/
Void create_cylinder( Tobject* obj, Tmatrix acc_trans )
{
#ifdef VERBOSE
  printf("creating a cylinder\n");	
  print_obj_info( obj );
	printf(" Radius is "PCF"\n", obj->spec.cyl.radius );
	printf(" Capped? Top %d Bottom %d \n ",
					obj->spec.cyl.tcapped,
					obj->spec.cyl.bcapped );
#endif

  _Warning("Ignoring MSDL cylinder");

#ifdef CLEANUP
  free( obj );
#endif
}
/*---------------------------------------------------------------------------*/

/*===========================================================================*/
Void create_cone( Tobject* obj, Tmatrix acc_trans )
{
#ifdef VERBOSE
  printf("Creating a cone\n");	
  print_obj_info( obj );
	printf(" Radii	Top: "PCF" Bottom: "PCF"\n",
					obj->spec.cone.tradius,
					obj->spec.cone.bradius );
	printf(" Capped? Top %d Bottom %d \n ",
					obj->spec.cone.tcapped,
					obj->spec.cone.bcapped );
#endif

  _Warning("Ignoring MSDL cone");

#ifdef CLEANUP
  free( obj );
#endif
}
/*---------------------------------------------------------------------------*/

/*===========================================================================*/
Void create_disc( Tobject* obj, Tmatrix acc_trans )
{
#ifdef VERBOSE
  printf("Creating a disc\n");	
  print_obj_info( obj );
	printf(" Centre is "PCF" "PCF" "PCF" \n",
						obj->spec.disc.centre.x,
						obj->spec.disc.centre.y,
						obj->spec.disc.centre.z );
	printf(" It's radius is "PCF"\n", obj->spec.disc.radius );
#endif

  /* Do with this object as you will but please be gentle!! */
  _Warning("Ignoring MSDL disc");

#ifdef CLEANUP
  free( obj );
#endif
}
/*---------------------------------------------------------------------------*/

/*===========================================================================*/
Void create_box( Tobject* obj, Tmatrix acc_trans )
{
#ifdef VERBOSE
  printf("Creating a box\n");	
  print_obj_info( obj );
	printf(" Top right is  "PCF" "PCF" "PCF" \n",
						obj->spec.box.tr.x,
						obj->spec.box.tr.y,
						obj->spec.box.tr.z );
	printf(" Bottom left is "PCF" "PCF" "PCF" \n",
						obj->spec.box.bl.x,
						obj->spec.box.bl.y,
						obj->spec.box.bl.z );
#endif
	Vec3d &p1 = (Vec3d&) obj->spec.box.tr;
	Vec3d &p2 = (Vec3d&) obj->spec.box.bl;
	Mat4d &t = (Mat4d&) acc_trans[0][0];
	Vector r;

	slObject("cube");
	FindMaxElts(p1 - p2, p2 - p1, r);
	slApply(xform(t, 
			xform(Shift((p1 + p2) / 2.0),
				  Scale(r))));

#ifdef CLEANUP
  free( obj );
#endif
}
/*---------------------------------------------------------------------------*/

/*===========================================================================*/
Void create_nurb( Tobject* obj, Tmatrix acc_trans )
{
#ifdef VERBOSE
  printf("Creating a nurb\n");	
  print_obj_info( obj );
#endif

  _Warning("Ignoring MSDL nurb");

#ifdef CLEANUP
  free( obj->spec.nurb.knots_u );
  free( obj->spec.nurb.knots_v );
  free( obj->spec.nurb.cpts    );
  free( obj );
#endif
}
/*---------------------------------------------------------------------------*/

/*===========================================================================*/
Void create_polyhedron( Tobject* obj, Tmatrix acc_trans )
{
#ifdef CLEANUP
	int 		i=0;		/* needed for CLEANUP */
#endif

#ifdef VERBOSE
	printf("Creating a polyhedron\n");
	print_obj_info( obj );
	printf(" It has %4d vertices \n",obj->spec.polh.vvnfn.no_verts );
	printf(" It has %4d vertex normals \n",obj->spec.polh.vvnfn.no_vnorms );
	printf(" It has %4d facet  normals \n",obj->spec.polh.vvnfn.no_fnorms );
#endif

	Int		i, j;
	Point	q;
	
	HandleProps(obj);
	slPointList();
	for (i = 0; i < obj->spec.tri.vvnfn.no_verts; i++)
		slPoint(toPoint(transPt3(obj->spec.tri.vvnfn.data + i, acc_trans)));

	slBeginFaces();
	for (j = 0; j < obj->spec.polh.flist.no_faces; j++)
	{
		Tsingleface *f = obj->spec.polh.flist.face[j];
		for (i = 0; i < f->no_verts; i++)
		{
			Assert(f->vert_index[i] > 0,
				   "0-based vert_index list");
			slPointIndex(f->vert_index[i] - 1);
		}
		slFace();
	}
	slEndFaces();

#ifdef CLEANUP
  free( obj->spec.polh.vvnfn.data );
  for( i=0 ; i<obj->spec.polh.flist.no_faces ; i++ )
    free( obj->spec.polh.flist.face[i]->vert_index );
  free( obj->spec.polh.flist.face );
  free( obj );
#endif
}
/*---------------------------------------------------------------------------*/

Void print_obj_info( Tobject* obj )
{ 
  /* this prints out info about an object a bit like the sdl command walk 
     but it executed in the application writers code... that's why it's here!! */

  printf ( " Obj Name: %s\n",  obj->name );
  printf ( "  Obj Type: %i\n",  obj->otype );
  PRINTP(obj->sprops);
  printf ( "  "PCF" "PCF" "PCF"\n", 
	  obj->box.min.x, obj->box.min.y, obj->box.min.z );
  printf ( "  "PCF" "PCF" "PCF"\n", 
	  obj->box.max.x, obj->box.max.y, obj->box.max.z );
}
/*****************************************************************************
	**	  **    ****    **    **  ********    *****
	**	  **  **    **  **    **     **     **     *
	**	  **  **  ****  ********     **       ****
	**	  **  **    **  **    **     **     *     **
	********  **    ****    **    **     **      *****
 *****************************************************************************/

Void create_point_lt( Tlights* lite )
{
#ifdef VERBOSE
  printf( "Creating point light of type %d\n",lite->ltype );
  printf( "at "PCF" "PCF" "PCF"\n",	lite->spec.point.locn.x,
	 				lite->spec.point.locn.y,
	 				lite->spec.point.locn.z );
  PRINTrgb( "Intensity = ", lite->spec.point.inten );
#endif

  _Warning("Ignoring MSDL point light");

#ifdef CLEANUP
  free( lite );
#endif
}

Void create_directional_lt( Tlights* lite )
{
#ifdef VERBOSE
  printf( "Creating directional light of type %d\n",lite->ltype );
  printf( "towards "PCF" "PCF" "PCF"\n",	lite->spec.direc.along.x,
	 				lite->spec.direc.along.y,
	 				lite->spec.direc.along.z );
  PRINTrgb( "Intensity = ", lite->spec.direc.inten );
#endif

  _Warning("Ignoring MSDL directional light");

#ifdef CLEANUP
  free( lite );
#endif
}  

Void create_spot_lt( Tlights* lite )
{
#ifdef VERBOSE
  printf( "Creating point light of type %d\n",lite->ltype );
  printf( "at "PCF" "PCF" "PCF"\n",	lite->spec.spot.locn.x,
	 				lite->spec.spot.locn.y,
	 				lite->spec.spot.locn.z );
  printf( "towards "PCF" "PCF" "PCF"\n",lite->spec.spot.at.x,
	 				lite->spec.spot.at.y,
	 				lite->spec.spot.at.z );
  PRINTrgb( "Intensity =" , lite->spec.spot.inten );
#endif

  _Warning("Ignoring MSDL spot light");

#ifdef CLEANUP
  free( lite );
#endif
}

/*---------------------------------------------------------------------------*/

/****************************************************************************
  *****  *    *  ******          *    *     *    ******  *    *
    *    *    *  *               *    *     *    *       *    *
    *    ******  *****           *    *     *    *****   * ** *
    *    *    *  *                *  *      *    *       **  **
    *    *    *  ******            **       *    ******  *    * 
 ***************************************************************************/
/*===========================================================================*/
Void create_view( Tview* view )
{
#ifdef VERBOSE
  printf( "Creating a view\n" );
  printf( " from:    "PCF" "PCF" "PCF"\n",	view->from.x,
						view->from.y,
						view->from.z );
  printf( " towards: "PCF" "PCF" "PCF"\n",	view->towards.x,
						view->towards.y,
						view->towards.z );
  printf( " vpnorm:  "PCF" "PCF" "PCF"\n",	view->vpn.x,
						view->vpn.y,
						view->vpn.z );
  printf( " up:      "PCF" "PCF" "PCF"\n",	view->up.x,
						view->up.y,
						view->up.z );
  printf( " window:  "PCF" "PCF" "PCF" "PCF"\n",view->window.bl.u,
						view->window.bl.v,
						view->window.tr.u,
						view->window.tr.v );
  printf( " front:   "PCF"\n",			view->fpd );
  printf( " back:    "PCF"\n",			view->bpd );
  printf( " persp:   %d\n",			view->persp );
#endif

  /* Do with this view as you will but please be gentle!! */
  _Warning("Ignoring MSDL view");

  gUp = (Vec3d&) view->up;

#ifdef CLEANUP
  free( view );
#endif
}

#endif
