/*
	File:		ParseObj.cc

	Function:	Read wavefront "obj" files

	Author:		Andrew Willmott

	Notes:		
*/

#include <string.h>
#include <unistd.h>
#include "gcl/Scene.h"
#include "gcl/SceneObjects.h"
#include "gcl/Avars.h"
#include "gcl/SceneLang.h"

scScenePtr ParseObjFile(const Char *filename);

static Void ReadIndexes(istream &s, Int &idx, Int &tidx, Int &nidx)
{
	Char	c;
	
	idx = tidx = nidx = 0;
	if (isdigit(s.peek()))
		s >> idx;
	if (s.peek() == '/')
	{
		s.get();
		if (isdigit(s.peek()))
			s >> tidx;
		if (s.peek() == '/')
		{
			s.get();
			if (isdigit(s.peek()))
				s >> nidx;
		}
	}

	idx--;
	nidx--;
	tidx--;
}

/*
	Example material file:
	
	newmtl Cube_0
		Ns 100
		d 1
		illum 2
		Ka .2 .2 .2
		Kd 0.8 0.8 0.8
		Ks 0 0 0
		map_Kd finishedx0001.jpg
		map_Ka finishedx0001.jpg
*/

struct Mtl
{
				Mtl() : flags(0) {};
	
	UInt32		flags;
	Colour		kd;
	String		texFile;	

	Void 		ApplyMtl();
};

enum MtlFlags 
{
	hasTexture = 1,
	hasKd = 2,
};

Void Mtl::ApplyMtl()
{
	if (flags & hasKd)
		slColour(kd);
	if (flags & hasTexture)
	{
		cerr << "setting texture " << texFile << endl;
		slEndFaces();
		slTexture(texFile);
		slBeginFaces();
	}
}

typedef Array<Mtl> MtlList;

MtlList gMtlList;
IntHash gMtlNames;
 
Int ParseMaterialFile(StrConst filename)
{
	String		token;
	ifstream 	s;
	Colour		c;

	cerr << "parsing material file " << filename << endl;
	s.open(filename);
	if (!s)
	{
		cerr << "(ParseMaterialFile) Cannot access " << filename << endl;
		return(-1);
	}

	while (s)
	{
		if (token.ReadWord(s))
		{
			if (token[0] == '#')
				;
			else if (token == "newmtl")
			{
				String mtlName;

				mtlName.ReadWord(s);
				cerr << "new material " << mtlName << endl;
				gMtlNames.SetItem(mtlName, gMtlList.NumItems());
				gMtlList.Add(1);
			}
			else if (token == "Kd")
			{
				s >> c[0] >> c[1] >> c[2];
				gMtlList.Last().kd = c;
				gMtlList.Last().flags |= hasKd;
			}
			else if (token == "map_Kd")
			{
				String texFile;
				
				texFile.ReadWord(s);
				SubstituteEnvVars(texFile);
				
				cerr << "texture file " << texFile << endl;
				gMtlList.Last().flags |= hasTexture;
				gMtlList.Last().texFile = texFile;
			}
			else
				cerr << "(ParseMaterialFile) *** Ignoring unknown token: " << token << endl;

			token.ReadLine(s); // ignore rest of the line
		}
	}
	return(0);
}

#define SMF_OLD

scScenePtr ParseObjFile(const Char *filename)
// Parses wavefront-style .obj file.
/*
	Format is: 'function args' on each line.
	
	v x y z				add vertex
	vt u v [w]			add texture vert.
	vn i j k			add normal
	f 1 5 3 2			define polygon
						indices are of the form v[/vt[/vn]] 
						indices are 1-based.
	g group1 group2... 	set group names for following f's.
	o					object name for following f's
	s n					set smoothing group.
	usemtl material		set material name for ...

	Extensions:

	c r g b				add colour
	
*/
{
	String		token;
	Point		vtx;
	Vector		nrm;
	Coord		tc;
	Colour		clr;
	Int			idx,  tidx, nidx, cidx, vtxOffset;
	Int			vertices, faces, colours, normals, texCoords;
	ifstream 	s;
	scScenePtr	result;

	// create default material
	gMtlList.Clear();
	gMtlList.Add(1);
	
	vtxOffset = -1; // for michael garland's SMF format

	s.open(filename);
	if (!s)
	{
		cerr << "(ParseObjFile) Cannot access " << filename << endl;
		return(0);
	}

	result = slBeginObject(filename);
	slPointList();
	slBeginFaces();
	vertices = colours = faces = normals = texCoords = 0;
	
	while (s)
	{
		if (token.ReadWord(s))
		{
			if (token[0] == '#')
				;
			else if (token == "begin" || token == "end")
			{
				// michael's stack demarcation? ignore for now
			}
			else if (token == "%SMF")
			{
				GCLReal version;
				s >> version;
				if (version < 1.0)
					vtxOffset = 0;
			}
#ifdef SMF_OLD
			else if (token == "t")
			{
				// SMF triangle : obsolete?
				s >> idx; slPointIndex(idx + vtxOffset);
				s >> idx; slPointIndex(idx + vtxOffset);
				s >> idx; slPointIndex(idx + vtxOffset);
				slFace();
				faces++;
			}
#endif			
			else if (token == "set")
			{
				token.ReadWord(s);
				if (token == "vertex_correction")
				{
					s >> vtxOffset;
					vtxOffset = 1 - vtxOffset;
				}
				else 
					cerr << "(ParseObjFile) *** Ignoring unknown set variable: "
						 << token << endl;
			}

			// standard wavefront stuff
			else if (token == "v")
			{
				s >> vtx[0] >> vtx[1] >> vtx[2];
				slPoint(vtx);
				vertices++;
			}
			else if (token == "vt")
			{
				s >> tc[0] >> tc[1];

				if (texCoords == 0)
					slCoordList();
				slCoord(tc);

				texCoords++;
			}
			else if (token == "vn" || token == "n")
			{
				s >> nrm[0] >> nrm[1] >> nrm[2];

				if (normals == 0)
					slNormalList();
				slNormal(nrm);

				normals++;
			}
			else if (token == "vc" || token == "c")	
			// extension to handle vertex colours
			{
				s >> clr[0] >> clr[1] >> clr[2];
				if (colours == 0)
					slColourList();
				slColour(clr);
				colours++;
			}
			else if (token == "g" || token == "s" || token == "o")
			{
				// ignore for now...
			}
			else if (token == "f" || token == "fo")
			{
				while(!IsEndOfLine(s))
				{
					ReadIndexes(s, idx, tidx, nidx);

					if (idx < 0) idx += vertices;
					if (tidx < 0) tidx += texCoords;
					if (nidx < 0) nidx += normals;

					slPointIndex(idx);

					if (tidx >= 0)
						slCoordIndex(tidx);
					if (nidx >= 0)
						slNormalIndex(nidx);
				}
				slFace();	
				faces++;
			}
			else if (token == "mtllib")
			{
				FileName	mtlFile;
				
				token.ReadWord(s);
				mtlFile.SetPath(filename);
				mtlFile.SetRelPath(token);
				ParseMaterialFile(mtlFile.GetPath());
			}
			else if (token == "usemtl")
			{
				String material;
				
				material.ReadWord(s);
				
				slEndFaces();

				if (gMtlNames.ItemExists(material))
				{
					// XXX
					Int mtlIdx = gMtlNames.GetItem(material);

					gMtlList[mtlIdx].ApplyMtl();
				}
				else if (material == "light")
				{
					slAttribute(new scAvarEmittance(Avar("light", 20, 5, 50), 
													cWhite));
					scColour(cBlack);
				}
				else if (material == "colour")
				{
					Colour mc;

					s >> mc[0] >> mc[1] >> mc[2];
					slColour(mc);  
				}
				else if (material == "white")
					slColour(cWhite);   
				else if (material == "black")
					slColour(cBlack);   
				else if (material == "red")
					slColour(cRed);   
				else if (material == "blue")
					slColour(cBlue);   
				else if (material == "green")
					slColour(cGreen);   
				else if (material == "grey")
					slAttribute(new scAvarColour(Avar("grey", 0.5, 0, 1),
												 cWhite));
				else
				{
					cerr << "(ParseObjFile) *** Warning: unrecognized material: " 
						 << material << endl;
					// set to default material
					gMtlNames.SetItem(material, 0);	
				}

				slBeginFaces();
			}
			else
				cerr << "(ParseObjFile) *** Ignoring unknown token: " << token << endl;

			token.ReadLine(s); // ignore rest of the line
		}
	}
	
	slEndFaces();
	slEndObject();
	s.close();
	
	cerr << "Read " << vertices << " vertices, ";
	if (colours) cerr << colours << " colours, ";
	if (normals) cerr << normals << " normals, ";
	if (texCoords) cerr << texCoords << " texture coordinates, ";
	cerr << "and " << faces << " faces." << endl;
	
	return(result);
}

#ifdef CL_SGI_INST
#pragma instantiate Array<Mtl>
#endif

