/*
	File:		Readers.cc

	Function:	Support for reading in various scene file formats

	Author:		Andrew Willmott

	Notes:		
*/

#include "gcl/Readers.h"

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include "gcl/GCLConfig.h"

#ifdef DEBUG
#define DBG_COUT if (1) cerr
#else
#define DBG_COUT if (0) cerr
#endif

#define INFO_COUT cerr

typedef scScenePtr (*SceneParser)(const Char *filename);

struct SceneParserEntry
{
	StrConst		extension;
	SceneParser		parser;
	StrConst		description;
};

scScenePtr ParseObjFile(const Char *filename);
scScenePtr ParseSLFile(const Char *filename);
scScenePtr ParseSCFile(const Char *filename);
scScenePtr ParseMGFFile(const Char *filename);
scScenePtr ParseMMFFile(const Char *filename);
scScenePtr ParseMSDLFile(const Char *filename);

static StrConst kSceneFileExtensions[] = 
{
	"sl",
	"obj",
	"msh",
	"smf",
	"mrb",
	"mmf",
	"cmf",
#ifdef GCL_PARSE_SC
	"sc",
#endif
#ifdef GCL_MGF
	"mgf",
#endif
#ifdef GCL_MSDL
	"msdl",
#endif
	0
};

static SceneParserEntry kSceneParserTable[] = 
{
	"sl",	ParseSLFile,		"GCL's scene language.",
	"obj",	ParseObjFile,		"Wavefront .obj text format",
	"msh",	ParseObjFile,		".msh file",
	"smf",	ParseObjFile,		"Garland's .smf format (obj-like)",
	"mrb",	ParseMMFFile,		"binary multi-res model format",
	"mmf",	ParseMMFFile,		"ascii multi-res model",
	"cmf",	ParseMMFFile,		"ascii face cluster model",
#ifdef GCL_PARSE_SC
	"sc",	ParseSCFile,		"Heckbert's .sc format",
#endif
#ifdef GCL_MGF
	"mgf",	ParseMGFFile,		"Ward et al.'s .mgf format",
#endif
#ifdef GCL_MSDL
	"msdl",	ParseMSDLFile,		"Manchester Scene Description Language",
#endif
	0
};

Void SceneReader::PrintSupportedFormats(ostream &s)
{
	SceneParserEntry *sp = kSceneParserTable;
	Int		i;

	s << "Supported scene file extensions: " << endl;

	while (sp->extension)
	{
		s << "  " << sp->extension;
		for (i = 0; i < 4 - strlen(sp->extension); i++)
			s << ' ';
		s << "    " << sp->description << endl;
		sp++;
	}
#ifdef CL_GZIP
	s << "gzipped versions of these files are also acceptable." << endl;
#endif
}

scScenePtr SceneReader::Load(FileName &sceneFile)
/*! 
	Sets sceneFile to the name of the scene file found, if
	extensions had to be added.

	Returns 0 if the scene could not be loaded.
*/
{
	scScenePtr			result;
	String				gunzipFile, origFile;
	Bool				gzipFile = false;
	SceneParserEntry	*sp;
	FileName			filename(sceneFile);
	Int					fileExt;

	fileExt = filename.FindFileExtension(kSceneFileExtensions);

	if (fileExt == kFileNotFound)
		INFO_COUT << filename.GetPath() << 
			" does not exist or is not readable." << endl;
	else if (fileExt == kBadExtension)
		INFO_COUT << filename.GetPath() << " has an unknown extension." 
				  << endl;

	if (fileExt < 0)
		return(0);
		
	sceneFile = filename;	// set sceneFile to the file we actually found

	if (filename.DecompressSetup() != 0)
	{
		INFO_COUT << "Error while decompressing " << origFile << endl;
		return(0);
	}
	INFO_COUT << "Reading " << filename.GetPath() << endl;
	result = (kSceneParserTable[fileExt].parser)(filename.GetPath());
	filename.DecompressCleanup();
	
	return(result);
}
