/*
	File:		SceneConvert.cc

	Function:	command-line program to convert scene files

	Author:		Andrew Willmott

	Notes:		
*/

#include <stdlib.h>

#include "arg.h"
#include "gcl/SceneLang.h"
#include "gcl/Readers.h"
#include "gcl/Writers.h"
#include "gcl/MRModel.h"
#include "gcl/Avars.h"

static StrConstArray	gAvarNames;
static ScalarList		gAvarVals;

static Int avarArg(Int argc, Char **argv)
{
	Int		i;
	GCLReal	val;
	
	if (argc % 2 == 1)
	{
		cerr << "Error: wrong number of avar arguments." << endl;
		exit(1);
	}
	
	for (i = 0; i < argc; i += 2)
	{
		gAvarNames.Append(argv[i]);
		val = atof(argv[i + 1]);
		gAvarVals.Append(val);
	}
	
	return(0);
}

Void SetAvars(scScenePtr scene)
{
	Int				i;
	CreateAvarList	makeAvarList;

	for (i = 0; i < gAvarNames.NumItems(); i++)
		cout << "avar " << gAvarNames[i] << " = " << gAvarVals[i] << endl;
		
	scene->ApplyAction(makeAvarList);
#ifdef DEBUG
	cout << "scene avars:" << endl;
	for (i = 0; i < makeAvarList.avarList->NumItems(); i++)
		cout << makeAvarList.avarList->Item(i).name << " = " 
			<< makeAvarList.avarList->Item(i).value << endl;
#endif

	scene->Set(makeAvarList.avarList);
	
	for (i = 0; i < gAvarNames.NumItems(); i++)
		if (!makeAvarList.avarList->SetAvar(gAvarNames[i], gAvarVals[i]))
			cerr << "No such avar in scene: " << gAvarNames[i] 
				 << " (ignoring) " << endl;
}

static Void PrintSceneInfo(scScenePtr scene)
{
	FindDecInfo	decInfo;
	
	scene->Decimate(decInfo, DecTris);
	
	cout << "polygons " << decInfo.numPolys << endl;
	cout << "triangles " << decInfo.numTris << endl;
}

Int main(Int argc, Char *argv[])
{
	FileName		inPath, outPath;
	Char			*file1 = 0, *file2 = 0;
	Int				formats, noRescale, writeCMB;
	scScenePtr		scene;
	Double			meshComp;
	arg_form		*arg_format;
	
	// command-line options
	
	meshComp = 1.0;
	
	arg_format = arg_to_form(0,
		"", 									"Usage: slconv [options]",
		
		"[%S]",	&file1,							"Input scene file",
		"[%S]",	&file2,							"Output scene file",
		"-set", ARG_SUBR(avarArg), 				"Set avar, e.g. -set light 0.5 height 0.2",
		"-meshComp %F", &meshComp, 				"Set mesh complexity for MR models",
		"-writeMRB", ARG_FLAG(&writeCMB),		"Write multires model(s) in binary format",
		"-noRescale", ARG_FLAG(&noRescale),		"Don't rescale the scene",
		"-formats", ARG_FLAG(&formats),			"List supported file formats",
		0);	

	if (arg_parse_argv(argc, argv, arg_format) < 0)
		return(1);

	if (formats)
	{
		cout << "Input formats:" << endl;
		SceneReader::PrintSupportedFormats(cout);
		cout << "Output formats:" << endl;
		SceneWriter::PrintSupportedFormats(cout);
		return(0);
	}

	if (!file1)
	{
		arg_form_print(arg_format);
		return(0);
	}
	
	inPath.SetPath(file1);
	if (file2)
		outPath.SetPath(file2);

	// normalise or not? 

	slInit();
	scMRModel::SetComplexity(meshComp);
	scene = SceneReader::Load(inPath);
	if (!scene)
		return(1);

	if (!noRescale)
		scene->Normalise();

	SetAvars(scene);

	if (writeCMB)
	{
		// write out models as binary
		MRModelsFinder	models;
		Int				i;
		
		scene->ApplyAction(models);

		for (i = 0; i < models.NumModels(); i++)
		{
			cout << "writing model " << models.Model(i)->modelFile.GetFile() << endl;
			models.Model(i)->WriteBinary();
		}
	}
	else
	if (!file2)
		// dump information
		PrintSceneInfo(scene);	
	else
		// write out scene
		SceneWriter::Save(scene, outPath);
	
	return(0);
}
