/*
	File:			ScenePane.cc

	Function:		See header file

	Author(s):		Andrew Willmott

	Copyright:		Copyright (c) 1995-1996, Andrew Willmott

	Notes:			May have to split some of this stuff out.
					To come: picking, quaternion rotation (mouseball) &
					mouse pole.
*/

#include "gcl/GCLConfig.h"

#ifndef GCL_NO_GL

#include "gcl/ScenePane.h"
#include "gcl/GCLConfig.h"
#include "gcl/MRModel.h"

#ifdef VL_ROW_ORIENT
#error not_implemented
// XXX Need to reorder rotation matrices. or use xform()
#endif


// --- Grid primitive ---------------------------------------------------------

// this is just for display -- don't add it to any model definitions.

class scBaseGrid : public scPrimitive
{
public:
	scBaseGrid() : scPrimitive(pExtension), divisions(10), on(false) {};
	
	Void	Draw(Renderer &r, SLContext *);
	
	Int		divisions;
	Bool	on;
};

Void scBaseGrid::Draw(Renderer &r, SLContext *)
{
	Int		i, j;
	GCLReal	x, y, d;
	Point	p1, p2, p3, p4;
	
	if (!on)
		return;
		
	d = 2.0 / divisions;
	
	r.C(Colour4(cGreen, 0.5));
	r.N(vl_y);
	p1 = p2 = p3 = p4 = vl_0;
	r.Begin(renLines);
	
	for (i = 0; i <= divisions; i++)
	{
		y = -1.0 + i * d;
		r.P(Point(-1.0, 0.0, y)).P(Point(1.0, 0.0, y));
		for (j = 0; j <= divisions; j++)
		{
			x = -1.0 + j * d;
			r.P(Point(x, 0.0, -1.0)).P(Point(x, 0.0, 1.0));
		}
	}
	r.End();

	r.C(Colour4(cCyan, 0.3));
	r.Begin(renPoly)
		.P(Point(-1.0, 0.0, -1.0))
		.P(Point(-1.0, 0.0,  1.0))
		.P(Point( 1.0, 0.0,  1.0))
		.P(Point( 1.0, 0.0, -1.0))
		.End();   
}


// --- ScenePane class --------------------------------------------------------


ScenePane::ScenePane(Bool doubleBuf) :
	GLXPane(doubleBuf), itsScene(0), itsCamera(0),
	rotX(0), rotY(0),
	transX(0), transY(0), transZ(0),
	zoom(1.0), 
	wire(false),
	gridShown(false),
	isSavedCam(false)
{
	SetBgColour(Colour(0.5, 0.5, 0.5));
}

Void ScenePane::SetScene(scScenePtr scene)
{
	itsCamera = (scCamera *) scene->FindFirst(aCamera);
	if (!itsCamera)
		itsCamera = new scCamera;

	itsScene = new scGroup;
	// add default colour
	itsScene->Add(new scColour(scPrimitive::sDefaultColour));
	// scene & camera
	itsScene->Add(itsCamera);
	itsScene->Add(scene);
	// add grid
	grid = new scBaseGrid;
	grid->on = gridShown;
	itsScene->Add(grid);
	
	SetCameraParams();

	// make a list of mr models in the scene
	scene->ApplyAction(models);
	
	Redraw();

	if (doubleBuffered)
		Redraw();
}

Void ScenePane::SetCameraParams()
{
	itsCamera->SetSceneOrient(
		Rotation(vl_x, rotX * vl_pi)
		* Rotation(vl_y, rotY * vl_pi));

	itsCamera->SetSceneOffset(Shift(Vector(transX, transY, transZ) * 4));
	itsCamera->SetZoom(zoom);
	itsCamera->SetAspect(GCLReal(width) / GCLReal(height));
}

Void ScenePane::PrintCameraParams(Bool animFormat)
{
	if (animFormat)
	{
		cout << "k cam_rot_c " << rotX << ' ' << rotY << endl;
		cout << "k cam_trans_v " << transX << ' ' << transY << ' ' << transZ << endl;
		if (zoom != 1.0)
			cout << "k zoom " << zoom << endl;
		cout << endl;
	}
	else
	{
		cout << "cam_rot_c " << rotX << ' ' << rotY << ' ';
		cout << "cam_trans_v " << transX << ' ' << transY << ' ' << transZ;
		if (zoom != 1.0)
			cout << " zoom " << zoom << endl;
		cout << endl;
	}
}

Void ScenePane::Redraw()
{
	Clear().Draw(itsScene).Show();
}

const GCLReal kElevLimit = 0.495; 
	// To stop us reaching the singularity on the y axis.


Void ScenePane::ViewFrom(Point &position, Vector &normal)
{
	itsCamera->SetPosition(position);
	itsCamera->SetLookat(position + normal);
	
	rotX = rotY = 0;
	transX = transY = transZ = 0;
	zoom = 1;

	SetCameraParams();
	
	Redraw();
	if (!isSavedCam)
	{
		savedCam = *itsCamera;
		isSavedCam = true;
	}
}

Void ScenePane::Restore()
{
	if (isSavedCam)
	{
		*itsCamera = savedCam;
		isSavedCam = false;
	}
}

// --- Scene dragging ---------------------------------------------------------

Void ScenePane::TrackMouse(Int sx, Int sy, Int modifiers)
{
	Int	 		x, y;
	Coord		start((sx + 0.5) / width, 1.0 - (sy + 0.5) / height);
	Coord		c = start;
	Int			newModifiers;

	StartDrag(start, modifiers);
	GetMouse(x, y, newModifiers);

	while (newModifiers & wmButtons)
	{
		c = Coord((x + 0.5) / width, 1.0 - (y + 0.5) / height);
		Drag(c, modifiers);
		GetMouse(x, y, newModifiers);
	} 
	
	EndDrag(c, modifiers);
}

Void ScenePane::StartDrag(Coord c, Int mod)
{
	tRotX = rotX;
	tRotY = rotY;
	tZoom = zoom;
	tTransX = transX;
	tTransY = transY; 
	tTransZ = transZ;

	dragStart = c;
	modifiersStart = mod;

	baseErr = scMRModel::GetComplexity();

	if (mod & (wmButton1 | wmButton2))
		if (baseErr > 0.5)
			scMRModel::SetComplexity(0.1);
}

Void ScenePane::Drag(Coord c, Int)
{
	// standard behaviour: button1 = rotation, button 2 = translation,
	// button 3 = adjust complexity.

	offset = c - dragStart;
	if (modifiersStart & wmButton1)
	{
		if (modifiersStart & wmControl)
			SceneTransDrag();
		else
			SceneRotDrag();
	}
	else if (modifiersStart & wmButton2)
		SceneTransDrag();
	else if (modifiersStart & wmButton3)
		 ComplexityDrag();
}

Void ScenePane::EndDrag(Coord, Int)
{
	rotX = tRotX;
	rotY = tRotY;
	transX = tTransX;
	transY = tTransY;
	transZ = tTransZ;
	zoom = tZoom;

	if (modifiersStart & (wmButton1 | wmButton2))
		if (baseErr != scMRModel::GetComplexity())
		{
			scMRModel::SetComplexity(baseErr);
			Redraw();
		}
}
	
Void ScenePane::ComplexityDrag()
{
	scMRModel::SetComplexity(Clip(baseErr + offset[1],
		(GCLReal) 0.0, (GCLReal) 1.0));

	Redraw();
}

Void ScenePane::SceneRotDrag()
{
    if (modifiersStart & wmShift)
	{
		tZoom = zoom - offset[1];
		itsCamera->SetZoom(tZoom);
	}
	else
	{
		tRotY = rotY + offset[0];	// rotation
		tRotX = rotX - offset[1];	// elevation
		if (tRotX < -kElevLimit)
			tRotX = -kElevLimit;
		else if (tRotX > kElevLimit)
			tRotX = kElevLimit;
	
		itsCamera->SetSceneOrient(
			Rotation(vl_x, tRotX * vl_pi)
			* Rotation(vl_y, tRotY * vl_pi));
	}
			
	Clear().Draw(itsScene).Show();
}

Void ScenePane::SceneTransDrag2D()
{
	if (modifiersStart & wmShift)
	{
		tZoom = zoom - offset[1];
		itsCamera->SetZoom(tZoom);
	}
	else
	{
		tTransX = transX + offset[0] / zoom;
		tTransY = transY + offset[1] / zoom;
		itsCamera->SetSceneOffset(Shift(Vector(tTransX, 
			tTransY, tTransZ) * 4));
	}

	Clear().Draw(itsScene).Show();
}

Void ScenePane::SceneTransDrag()
{
	if (modifiersStart & wmShift)
		tTransZ = transZ - offset[1];
	else
	{
		tTransX = transX + offset[0];
		tTransY = transY + offset[1];
	}
	itsCamera->SetSceneOffset(Shift(Vector(tTransX, tTransY, tTransZ) * 4.0));

	Clear().Draw(itsScene).Show();
}

Void ScenePane::HandleExpose()
{
	Redraw();
}

Void ScenePane::HandleKey(Char c)
{
	Int		i;
	
	switch (c)
	{
	case 'w':
		wire = !wire;
		if (wire)
			scPrimitive::sRenderStyle = renLineLoop;
		else
			scPrimitive::sRenderStyle = renPoly;
		Redraw();
		break;

	case 'G':
		SetGrid(!gridShown);
		Redraw();
		break;

	case 'S':
		for (i = 0; i < models.NumModels(); i++)
			models.Model(i)->SimplestModel();
		scMRModel::SetComplexity(0.0);
		Redraw();
		break;

	case 'C':
		for (i = 0; i < models.NumModels(); i++)
			models.Model(i)->MostComplexModel();
		scMRModel::SetComplexity(1.0);
		Redraw();
		break;

	case 'V':
		for (i = 0; i < models.NumModels(); i++)
			models.Model(i)->showMeta = !models.Model(i)->showMeta;
		Redraw();
		break;

	case 'F':
		for (i = 0; i < models.NumModels(); i++)
			models.Model(i)->showFaces = !models.Model(i)->showFaces;
		Redraw();
		break;

	case 'X':
		for (i = 0; i < models.NumModels(); i++)
			models.Model(i)->colourFaces = !models.Model(i)->colourFaces;
		Redraw();
		break;

	case '[':
		models.Model(0)->AdaptClusters(models.Model(0)->currentClusters - 1);
		Redraw();
		break;
	case ']':
		models.Model(0)->AdaptClusters(models.Model(0)->currentClusters + 1);
		Redraw();
		break;
		
	case '0':
		SetHeadlight(false);
		Redraw();
		break;
			
	case '1':
		SetHeadlight(true);
		Redraw();
		break;

	case '!':
		SetDoubleSided(true);
		Redraw();
		break;
			
	case ')':
		SetDoubleSided(false);
		Redraw();
		break;

	default:
		XEventPane::HandleKey(c);
	}
}

Void ScenePane::SetGrid(Bool on)
{
	gridShown = on;
	grid->on = on;
}

// --- ScenePane2D methods ----------------------------------------------------

Void ScenePane2D::Drag(Coord c, Int)
{
	offset = c - dragStart;
	SceneTransDrag2D();
}

#endif
