/*
	File:			Viewer.cc

	Function:		Generic viewer for gcl library

	Author(s):		Andrew Willmott

	Copyright:		Copyright (c) 1995-1999, Andrew Willmott

	Notes:			
*/

#include "arg.h"

#include "gcl/GCLApp.h"
#include "gcl/ScenePane.h"
#include "gcl/EPSRenderer.h"
#include "gcl/MRModel.h"
#include "gcl/Readers.h"
#include "gcl/Writers.h"
#include "cl/Timer.h"
#ifdef GCL_MESA
#include "gcl/MesaRenderer.h"
#endif

#include <unistd.h>

#define GCL_WATCH_FILE

#define DBG_COUT if (0) cerr

// XXX generalise & move to Image::DownSample or something

Void DownSample(Image &in, Image &out)
/// quick hack for now
{
	Int		i, j;
	static ColourList span1, span2, outSpan;

	out.SetSize(in.Width() / 2, in.Height() / 2);	
	span1.SetSize(in.Width());
	span2.SetSize(in.Width());
	outSpan.SetSize(out.Width());

	for (i = 0; i < out.Height(); i++)
	{
		in.GetSpan(2 * i, 0, in.Width(), span1.Ref());
		in.GetSpan(2 * i + 1, 0, in.Width(), span2.Ref());

		for (j = 0; j < out.Width(); j++)
		{
			outSpan[j]  = span1[2 * j] + span1[2 * j + 1];
			outSpan[j] += span2[2 * j] + span2[2 * j + 1];
			outSpan[j] /= 4.0;
		}

		out.SetSpan(i, 0, out.Width(), outSpan.Ref());
	}
}


// --- Viewer Pane ------------------------------------------------------------


enum SDCmd 
{
	sdNone,
	sdHelp,
	sdReload,
	sdNext,
	sdPrev,
	sdDumpEPS, 
	sdDumpPic,
	sdDumpHQPic,
	sdDumpScene,
	sdDumpTree,
	sdDumpCMB,
	sdQuit
};

class DrawScenePane : public ScenePane
{
public:
	DrawScenePane();

	Void			HandleKey(Char c); 
	Void			PrintSceneInfo();
	Void			TimeFrame();

	SDCmd			command;
};
	
DrawScenePane::DrawScenePane() : 
	command(sdNone), ScenePane()
{
}

Void DrawScenePane::HandleKey(Char c)
{
	Bool finished = true;

	switch (c)
	{
	case 'h':
	case '?':
		command = sdHelp;
		break;
	case 'e':
		command = sdDumpEPS;
		break;
	case 'd':
		command = sdDumpPic;
		break;
#ifdef GCL_MESA
	case 'D':
		command = sdDumpHQPic;
		break;
#endif
	case 's':
		command = sdDumpScene;
		break;
	case '|':
		command = sdDumpTree;
		break;
	case 'B':
		command = sdDumpCMB;
		break;
	case 'r':
		command = sdReload;
		break;
	case ',':
	case 0x1D:
		command = sdPrev;
		break;
	case ' ':
	case '.':
	case 0x1C:
		command = sdNext;
		break;
	case 'q':
	case 0x1B:
		command = sdQuit;
		break;
	case 'i':
		PrintSceneInfo();
		finished = false;
		break;
	case 't':
		TimeFrame();
		finished = false;
		break;
	case 'c':
		PrintCameraParams(false);
		finished = false;
		break;
	case 'k':
		PrintCameraParams(true);
		finished = false;
		break;
	case 'R':
		// print radiance camera params
		{
			HVector lookDir;
			Point	lookFrom;

			Transform ict = inv(ItsCamera()->ModelMatrix());

			lookFrom = proj(ict * HPoint(0.0, 0.0, 0.0, 1.0));
			lookDir = ict * HVector(0.0, 0.0, -1.0, 0.0);

			printf("rview -vtv -vp %g %g %g -vd %g %g %g -vu 0 1 0 -vh 60 -vv 60 -vo 0 -va 0 -vs 0 -vl 0\n",
				lookFrom[0], lookFrom[1], lookFrom[2], 
				lookDir[0], lookDir[1], lookDir[2]
			);
		}
		finished = false;
		break;
									
	default:
		ScenePane::HandleKey(c);
		finished = false;
	}

	if (finished)
		xgs->SignalDone();
}

Void DrawScenePane::PrintSceneInfo()
{
	Int			i;
	FindDecInfo	decInfo;
	
	ItsScene()->Decimate(decInfo, DecTris);
	
	cout << "Scene info:" << endl;
	cout << "  " << decInfo.numTris << " triangles" << endl;
	cout << "  " << decInfo.numPolys << " polygons" << endl;

	for (i = 0; i < models.NumModels(); i++)
	{
		if (models.Model(i)->clusters.NumItems() > 0)
		{
			cout << "  Model " << i << " has " << models.Model(i)->currentClusters
				 << " clusters out of " << models.Model(i)->faces.NumItems()
				 << ", and " << models.Model(i)->rootClusters.NumItems() << " root clusters" << endl;
		}
		if (models.Model(i)->vertices.NumItems() > 0)
			cout << "  Model " << i << " has " << models.Model(i)->currentFaces
				 << " faces out of " << models.Model(i)->faces.NumItems() << endl;
	}
}

Void DrawScenePane::TimeFrame()
{
	ProgramTimer 	timer;
	Float			time;
	Int				i, count = 100;
	
	cout << "timing redraw rate" << endl;
	timer.StartTimer();
	cout << timer.GetTimer() << endl;

	for (i = 0; i < count; i++)
	{
		Clear().Draw(ItsScene()).Show();
	}

	cout << timer.GetTimer() << endl;
	time = timer.GetTimer();

	timer.StartTimer();
	cout << timer.GetTimer() << endl;
	for (i = 0; i < count; i++)
		;
	cout << timer.GetTimer() << endl;
	time -= timer.GetTimer();
	time /= count;
	
	cout << "frame drawing time: " << time << " seconds" << endl;
	cout << "frames per second: " << 1.0 / time  << endl;
}


// --- ViewerApp --------------------------------------------------------------


class ViewerApp : public GCLApp
{
public:

	Void			PrintKeys();
	Void			SetOptions(Int argc, Char **argv);
	Int				Run();
	Void			DrawScene();
	Void			SetupModels(scScenePtr scene);
	
// options
	Char			*output;
	Int				window;
	Int				ord;
	Int				noFork;
	Double			clusComp;
	Double			meshComp;
	Int				noLight;
	
// fields
	XGraphicsSystem	*xgs;
	DrawScenePane	*displayPane;
};

Void ViewerApp::PrintKeys()
{
	cout << "Viewer control:" << endl
		 << "  mouse 1 -- rotation" << endl
		 << "  shift+mouse 1 -- zoom" << endl
		 << "  mouse 2 -- translate x & y" << endl
		 << "  shift+mouse 2 -- translate z" << endl
		 << "  mouse 3 -- adjust scene complexity" << endl
		 << endl
		 << "  . space - move on to next scene file" << endl
		 << "  , - move back to previous scene file" << endl
		 << endl
		 << "  w - toggle wireframe" << endl
		 << "  0 - turn off lighting" << endl
		 << "  1 - turn on lighting" << endl
		 << "  ) - single-sided polygons" << endl
		 << "  ! - double-sided polygons" << endl
		 << endl
		 << "  G - toggle grid" << endl
		 << "  S - simplest model" << endl
		 << "  C - complex model" << endl
		 << "  F - toggle display of faces" << endl
		 << "  V - toggle visualisation of vertices/face clusters" << endl
		 << "  X - toggle colouring of face clusters" << endl
		 << endl
		 << "  d - save image" << endl
		 << "  D - save super-sampled image" << endl
		 << "  s - save scene file" << endl
		 << "  e - save eps file" << endl
		 << "  i - print scene information" << endl
		 << "  c - save camera parameters" << endl
		 << "  q esc - quit" << endl
		 << endl;
}

Void ViewerApp::SetOptions(Int argc, Char **argv)
{
	arg_form	*arg_format, *argForm2;
	Double		grey, clr[3], defClr[3];
	Int			size;
	Int			formats;
	Int			keys;
	Int			avgClus;

	size = 400;
	xsize = -1;
	ysize = -1;
	output = 0;
	grey = 0.5;
	clr[0] = -1;
	defClr[0] = -1;
	clusComp = -1.0;
	meshComp = -1.0;
	
	arg_format = arg_to_form(0,
		"", 
			"Usage: slview file1 file2 ... [options]",
		"", ARG_SUBR(GetFileArgs), 
			"",
		"-o %S", &output,
			"Set output file",
		"-bgr %F", &grey,
			"Set background grey level",
		"-bgc %F %F %F", clr, clr + 1, clr + 2,	
			"Set colour background",
		"-defClr %F %F %F", defClr, defClr + 1, defClr + 2,	
			"Set default model colour",
		"-noFork", ARG_FLAG(&noFork),
			"Don't fork into background.",
		"-avgClus", ARG_FLAG(&avgClus),
			"Colour parent clusters according to average of children.",
		"-set", ARG_SUBR(GetAvarArgs),
			"Set avar, e.g. -set light 0.5 height 0.2",
		"-clusComp %F", &clusComp,
			"Set face cluster complexity",
		"-meshComp %F", &meshComp,
			"Set model complexity",
		"-noLight", ARG_FLAG(&noLight),
			"Don't use headlight",
		"-controls", ARG_FLAG(&keys),
			"List keyboard and mouse controls",
		"-formats", ARG_FLAG(&formats),	
			"List supported file formats",

		// Output options		
		"-size %d", &size,
			"Set window size",
		"-xysize %d %d", &xsize, &ysize,
			"Set output size",
		0);

	argForm2 = arg_to_form(0,
			"-noScale", ARG_FLAG(&noScale),
			"Don't rescale scene.",
			"-zup", ARG_FLAG(&zUp),
				"z axis is up (default is y axis.)",
			"-zdown", ARG_FLAG(&zDown),
				"z axis is down",
		0);
	
	arg_form_append(arg_format, argForm2);

	if (argc == 1)
	{
		arg_form_print(arg_format);
		cout << endl;
		exit(0);
	}

	if (arg_parse_argv(argc, argv, arg_format) < 0)
		exit(1);

#ifdef DEBUG
	noFork = true;
#endif

	if (formats)
	{
		SceneReader::PrintSupportedFormats(cout);
		exit(0);
	}

	if (keys)
	{
		PrintKeys();
		exit(0);
	}		

	if (!avgClus)
		MRModel::sAvgClusColours = false;
	
	if (clr[0] < 0)
		bgColour = cWhite * grey;
	else
		bgColour = Colour(clr[0], clr[1], clr[2]);

	if (defClr[0] >= 0)
		scPrimitive::sDefaultColour = Colour(defClr[0], defClr[1], defClr[2]);

	if (xsize < 0)
		xsize = size;
	if (ysize < 0)
		ysize = size;
}

Int ViewerApp::Run()
{
	if (numFiles < 1)
		return(1);
		
	xgs = new XGraphicsSystem;
	slInit();

	if (!noFork && fork())
		return(0);

	DrawScene();

	return(0);
}

Void ViewerApp::SetupModels(scScenePtr scene)
{
	Int		i;

	if (meshComp >= 0.0)
		scMRModel::SetComplexity(meshComp);	
	else if (clusComp >= 0.0)
	{
		MRModelsFinder	models;

		scene->ApplyAction(models);

		scMRModel::SetComplexity(clusComp);

		for (i = 0; i < models.NumModels(); i++)
		{
			models.Model(i)->showFaces = false;
			models.Model(i)->showMeta = true;
			models.Model(i)->AdaptComplexity(clusComp);
		}
	}
}

Void ViewerApp::DrawScene()
{
	Int						i, j;
	RGBAImage				image;
	FileName				sceneFile, outFile;
	ofstream				fout;
	Bool					window = false, reload;
	
	scene = 0;
	displayPane = new DrawScenePane();
	displayPane->SetBgColour(bgColour);
	itsScenePane = displayPane;
		
	i = 0;
	reload = true;
	
	while (displayPane->command != sdQuit)
	{
		if (reload)
		{
			delete scene;
			sceneFile.SetPath(files[i]);

			ReadScene(sceneFile);
			if (!scene)
				exit(1);
			
			SetAvars(scene);
			SetupModels(scene);

			if (output)
				outFile.SetPath(output);
			else
			{
				outFile = sceneFile;
				outFile.SetDir(".");
				outFile.SetExtension("");
			}
			sceneFile.AddCompExt();	// add the compression extension so we
									// can watch the file
		}
		
		// we do this here to avoid creating a window until we know we've
		// successfully read the file.
		if (!window)
		{
			xgs->CreateWindow(displayPane, files[0],
							xsize, ysize);
			displayPane->SetHeadlight(!noLight);
			window = true;
		}

		if (reload)
		{			
			displayPane->SetScene(scene);
			reload = false;
		}
		
		// spin while the user manipulates the scene

#ifndef GCL_WATCH_FILE
		xgs->Run();
#else
		Long			timeStamp;

		displayPane->command = sdNone;
		timeStamp = sceneFile.GetTimeStamp();

		while (true)
		{
			xgs->RunFor(1.0);	// check once a second
			if (displayPane->command != sdNone)
				break;

			if (sceneFile.GetTimeStamp() > timeStamp)
			{
				displayPane->command = sdReload;
				break;
			}
		}
#endif

		switch (displayPane->command)
		{
		case sdQuit:
			break;
		case sdHelp:
			PrintKeys();
			break;
		case sdReload:
			reload = true;
			break;
		case sdPrev:
			if (i > 0)
				i--;
			else
				i = numFiles - 1;
			reload = true;
			break;
		case sdNext:
			i++;
			if (i == numFiles)
				i = 0;
			reload = true;
			break;
		case sdDumpPic:
			if (outFile.GetExtension() == "")
				outFile.SetExtension("tif");
			outFile.MakeUnique();
			cout << "*** saving image to " << outFile.GetPath() << endl;
			displayPane->GetImage(image);
			image.Save(outFile);
			break;
#ifdef GCL_MESA
		case sdDumpHQPic:
			{
				MesaRenderer 	mr;
				RGBAImage		hqImage;

				if (outFile.GetExtension() == "")
					outFile.SetExtension("tif");
				outFile.MakeUnique();
				cout << "*** saving high quality image to " << outFile.GetPath() << endl;
				
				image.SetSize(2 * xsize, 2 * ysize);
				mr.Init(image);
				mr.SetHeadlight(!noLight);
				mr.SetBgColour(bgColour);
				mr.Clear().Draw(displayPane->ItsScene()).Show();
				DownSample(image, hqImage);
				hqImage.Save(outFile);
			}
			break;
#endif
		case sdDumpScene:
			cout << "*** saving scene to " << outFile.GetPath() << endl;
			SceneWriter().Save(displayPane->ItsScene(), outFile);
			break;
		case sdDumpTree:
			outFile.SetExtension("hier");
			cout << "*** saving .hier file to " << outFile.GetPath() << endl;

			if (displayPane->models.NumModels() > 0)
				displayPane->models.Model(0)->DumpFCH(outFile.GetPath());
			break;
		case sdDumpCMB:
			if (displayPane->models.NumModels() > 0)
				displayPane->models.Model(0)->WriteBinary();
			break;
		case sdDumpEPS:
			EPSRenderer		eps;

			outFile.SetExtension("eps");
			cout << "*** saving EPS file to " << outFile.GetPath() << endl;
			eps.itsPlot.level2 = true;
			eps.Attach(outFile.GetPath());
			eps.Clear().Draw(displayPane->ItsScene()).Show();
			break;
		}
	}
}

// ----------------------------------------------------------------------------

main(Int argc, Char **argv)
{
	ViewerApp	app;
	
	app.SetOptions(argc, argv);

	return(app.Run());
}
